# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from datetime import datetime
from typing import Optional

import pandas as pd
from ax.adapter.base import Adapter
from ax.adapter.registry import Generators

from ax.analysis.analysis_card import ErrorAnalysisCard
from ax.analysis.overview import OverviewAnalysis
from ax.analysis.plotly.arm_effects import ArmEffectsPlot
from ax.analysis.plotly.scatter import ScatterPlot
from ax.analysis.results import ResultsAnalysis
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig

from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.metric import Metric
from ax.core.optimization_config import Objective, OptimizationConfig
from ax.core.outcome_constraint import ScalarizedOutcomeConstraint
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.core.search_space import SearchSpace
from ax.generation_strategy.generation_strategy import (
    GenerationNode,
    GenerationStrategy,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import MinTrials
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_data,
    get_experiment_with_scalarized_objective_and_outcome_constraint,
    get_offline_experiments,
    get_online_experiments,
)
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_default_generation_strategy_at_MBM_node


class TestOverview(TestCase):
    @mock_botorch_optimize
    def test_compute(self) -> None:
        # Set up a basic optimization with the Client
        for use_constraint in (False, True):
            client = Client()

            client.configure_experiment(
                name="booth_function",
                parameters=[
                    RangeParameterConfig(
                        name="x1",
                        bounds=(-10.0, 10.0),
                        parameter_type="float",
                    ),
                    RangeParameterConfig(
                        name="x2",
                        bounds=(-10.0, 10.0),
                        parameter_type="float",
                    ),
                ],
            )
            client.configure_optimization(
                objective="-1 * booth",
                outcome_constraints=["foo >= 0.0"] if use_constraint else None,
            )
            # Iterate well into the BO phase
            for _ in range(10):
                for trial_index, parameters in client.get_next_trials(
                    max_trials=1
                ).items():
                    raw_data = {
                        # pyre-ignore[58]
                        "booth": (parameters["x1"] + 2.0 * parameters["x2"] - 7) ** 2.0
                        # pyre-ignore[58]
                        + (2.0 * parameters["x1"] + parameters["x2"] - 5.0) ** 2.0,
                    }
                    if use_constraint:
                        raw_data["foo"] = float(parameters["x1"])  # pyre-ignore[58]
                    client.complete_trial(trial_index=trial_index, raw_data=raw_data)

            # Add a CANDIDATE batch trial to produce some trial analysis cards
            client._experiment.new_batch_trial()

            # Add metric fetching errors to produce some healthcheck analyses
            client._experiment._metric_fetching_errors = {
                (0, "booth"): {
                    "trial_index": 0,
                    "metric_name": "booth",
                    "reason": "This is a test",
                    "timestamp": datetime.now().isoformat(),
                    "traceback": "Test traceback",
                }
            }

            overview_card = OverviewAnalysis().compute(
                experiment=client._experiment,
                generation_strategy=client._generation_strategy,
            )
            self.assertEqual(
                any(
                    card.name == "ObjectivePFeasibleFrontierPlot"
                    for card_group in overview_card.children
                    for card in card_group.flatten()
                ),
                use_constraint,
            )

            children_names = [child.name for child in overview_card.children]
            self.assertIn("ResultsAnalysis", children_names)
            self.assertIn("InsightsAnalysis", children_names)
            self.assertIn("DiagnosticAnalysis", children_names)
            self.assertIn("AllTrialsAnalysis", children_names)
            self.assertIn("HealthchecksAnalysis", children_names)

            # test without BotorchGenerator
            overview_card = OverviewAnalysis().compute(
                experiment=client._experiment,
                # omit generation_strategy and pass adapter instead
                adapter=Generators.SOBOL(
                    experiment=client._experiment,
                    search_space=client._experiment.search_space,
                ),
            )
            # ObjectivePFeasibleFrontierPlot should not be computed
            # without a BoTorchGenerator
            self.assertFalse(
                any(
                    card.name == "ObjectivePFeasibleFrontierPlot"
                    for card_group in overview_card.children
                    for card in card_group.flatten()
                )
            )

    @mock_botorch_optimize
    def test_online(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an online setting.
        try:
            import pymoo  # noqa: F401

            expect_error_for_obj_pfeasible = False
        except ImportError:
            expect_error_for_obj_pfeasible = True
        analysis = OverviewAnalysis()

        for experiment in get_online_experiments():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            card_group = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )
            # we expect the objective vs pfeasible plot to error out if it is an
            # experiment with one objective and constranits and pymoo is not
            # installed.
            total_errors = sum(
                isinstance(card, ErrorAnalysisCard) for card in card_group.flatten()
            )
            self.assertEqual(
                total_errors,
                1
                if (expect_error_for_obj_pfeasible and not experiment.is_moo_problem)
                else 0,
            )
            for card in card_group.flatten():
                # TODO: add more AnalysisCard types when they support relativization
                if isinstance(card, (ArmEffectsPlot, ScatterPlot)):
                    self.assertIn("Relativized", card.title)

    @mock_botorch_optimize
    def test_offline(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        analysis = OverviewAnalysis()

        for experiment in get_offline_experiments():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            card_group = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )
            for card in card_group.flatten():
                self.assertNotIsInstance(card, ErrorAnalysisCard)
                self.assertNotIn("Relativized", card.title)

    def test_bandit_experiment_dispatch(self) -> None:
        experiment = Experiment(
            name="bandit_test",
            search_space=SearchSpace(
                parameters=[
                    ChoiceParameter(
                        name="x1",
                        parameter_type=ParameterType.FLOAT,
                        values=[-10.0, -7.5, -5.0, -2.5, 0.0, 2.5, 5.0, 7.5, 10.0],
                    ),
                    ChoiceParameter(
                        name="x2",
                        parameter_type=ParameterType.FLOAT,
                        values=[-10.0, -7.5, -5.0, -2.5, 0.0, 2.5, 5.0, 7.5, 10.0],
                    ),
                ]
            ),
            optimization_config=OptimizationConfig(
                objective=Objective(
                    metric=Metric(name="booth"),
                    minimize=True,
                )
            ),
        )

        # Create multi-arm trials to enable bandit analysis
        arm_configs = [
            [(-10.0, -10.0), (0.0, 0.0), (10.0, 10.0)],
            [(-10.0, 10.0), (10.0, -10.0), (5.0, 5.0)],
        ]

        trials = []
        data_rows = []

        for arm_coords in arm_configs:
            # Create batch trials with arms
            arms = [Arm(parameters={"x1": x1, "x2": x2}) for x1, x2 in arm_coords]
            trial = experiment.new_batch_trial()
            trial.add_arms_and_weights(arms=arms)
            trials.append(trial)

            # Generate data rows in same loop
            for arm in trial.arms:
                x1, x2 = float(arm.parameters["x1"]), float(arm.parameters["x2"])
                data_rows.append(
                    {
                        "trial_index": trial.index,
                        "arm_name": arm.name,
                        "metric_name": "booth",
                        "mean": (x1 + 2 * x2 - 7) ** 2 + (2 * x1 + x2 - 5) ** 2,
                        "sem": 0.0,
                    }
                )

        data = Data(df=pd.DataFrame(data_rows))

        # Attach data to the experiment
        experiment.attach_data(data)

        factorial_gen_node = GenerationNode(
            node_name="FACTORIAL",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.FACTORIAL,
                )
            ],
            transition_criteria=[
                MinTrials(
                    threshold=1,
                    transition_to="EMPIRICAL_BAYES_THOMPSON_SAMPLING",
                )
            ],
        )

        eb_ts_gen_node = GenerationNode(
            node_name="EMPIRICAL_BAYES_THOMPSON_SAMPLING",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.EMPIRICAL_BAYES_THOMPSON,
                )
            ],
            transition_criteria=None,
        )

        # Setup Bandit Generation Strategy
        bandit_gs = GenerationStrategy(
            name=Keys.FACTORIAL_PLUS_EMPIRICAL_BAYES_THOMPSON_SAMPLING,
            nodes=[factorial_gen_node, eb_ts_gen_node],
        )

        # Set the current node to the empirical bayes thompson node as that is
        # what is used to generate the marginal effects plot
        bandit_gs._curr = bandit_gs._nodes[1]

        # Compute the overview card
        card = OverviewAnalysis().compute(
            experiment=experiment,
            generation_strategy=bandit_gs,
        )

        # Flatten all cards to check for bandit-specific analyses
        all_cards = card.flatten()
        card_names = [c.name for c in all_cards]

        # Check that bandit-specific analyses are included
        self.assertIn("BanditRollout", card_names)
        self.assertIn("MarginalEffectsPlot", card_names)

        # Check that cross validation plots and top surfaces are excluded for bandit
        # experiments
        self.assertNotIn("CrossValidationPlot", card_names)
        self.assertNotIn("TopSurfacesAnalysis", card_names)

        # Ensure BanditRollout card is not an error card
        bandit_rollout_cards = [c for c in all_cards if c.name == "BanditRollout"]
        self.assertEqual(len(bandit_rollout_cards), 1)
        self.assertNotIsInstance(bandit_rollout_cards[0], ErrorAnalysisCard)

        marginal_effects_cards = [
            c for c in all_cards if c.name == "MarginalEffectsPlot"
        ]

        # Since there might be multiple MarginalEffectsPlot cards, check that at
        # least one exists and that none of them are error cards
        self.assertGreaterEqual(len(marginal_effects_cards), 1)

        # Check if adapter has predict properly implemented
        def has_predict_implemented(adapter: Optional[Adapter]) -> bool:
            """Check if adapter has predict method properly implemented."""
            if adapter is None or not hasattr(adapter, "predict"):
                return False
            try:
                adapter.predict(observation_features=[])
                return True
            except NotImplementedError:
                return False
            except Exception:
                # Other exceptions suggest method is implemented but needs proper setup
                return True

        predict_implemented = has_predict_implemented(bandit_gs.adapter)

        if predict_implemented:
            # Check that the marginal effects cards are not error cards
            for card in marginal_effects_cards:
                self.assertNotIsInstance(card, ErrorAnalysisCard)

    @mock_botorch_optimize
    def test_results_analysis_with_scalarized_constraints(self) -> None:
        """Test ResultsAnalysis works correctly with ScalarizedOutcomeConstraints."""
        # Create an experiment specifically with scalarized outcome constraints
        experiment = get_experiment_with_scalarized_objective_and_outcome_constraint()
        trial = experiment.new_batch_trial(
            generator_run=GeneratorRun(
                # The arm needs to be satisfying the parameter constraints
                # to be included in modeling
                arms=[
                    Arm(parameters={"w": 5.1, "x": 5, "y": "foo", "z": True, "d": 11.2})
                ]
            ),
            should_add_status_quo_arm=True,
        )
        trial.mark_running(no_runner_required=True)

        data = []
        for m_name, _ in experiment.metrics.items():
            data.append(
                get_data(
                    metric_name=m_name,
                    trial_index=trial.index,
                    num_non_sq_arms=1,
                    include_sq=True,
                )
            )
        data = Data.from_multiple_data(data)
        experiment.attach_data(data)
        trial.mark_completed()

        # Create a generation strategy for the test
        generation_strategy = get_default_generation_strategy_at_MBM_node(
            experiment=experiment
        )

        # Test that ResultsAnalysis doesn't error with scalarized constraints
        analysis = ResultsAnalysis()
        card_group = analysis.compute(
            experiment=experiment,
            generation_strategy=generation_strategy,
        )

        # Should not error and should produce a valid card group
        self.assertIsNotNone(card_group)
        self.assertEqual(card_group.title, "Results Analysis")

        # Should have some children (at least one analysis card)
        self.assertGreater(len(card_group.children), 0)

        # No error cards should be present
        for card in card_group.flatten():
            self.assertNotIsInstance(card, ErrorAnalysisCard)

        # Verify the experiment actually has scalarized outcome constraints
        self.assertIsNotNone(experiment.optimization_config)
        outcome_constraints = experiment.optimization_config.outcome_constraints
        self.assertIsNotNone(outcome_constraints)

        scalarized_constraints = [
            constraint
            for constraint in outcome_constraints
            if isinstance(constraint, ScalarizedOutcomeConstraint)
        ]
        self.assertGreater(
            len(scalarized_constraints),
            0,
            "Experiment should have at least one ScalarizedOutcomeConstraint",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from collections.abc import Mapping
from dataclasses import dataclass

import numpy as np
import torch
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.core.types import TParamValue
from torch import Tensor
from xgboost import XGBRegressor


@dataclass(kw_only=True)
class HierarchicalSearchSpaceSurrogate(BenchmarkTestFunction):
    lst_active_param_names: list[list[str]]
    flag_param_names: list[str]
    lst_flag_config: list[dict[str, bool]]
    lst_xgb_models: list[XGBRegressor]

    def evaluate_true(self, params: Mapping[str, TParamValue]) -> Tensor:
        """
        This function does the following two things to compute the test function values:
        1. Parse the flag parameters to decide which model to use;
        2. Collect all active parameters and pass them to the selected model.
        """
        flag_config = {
            key: value for key, value in params.items() if key in self.flag_param_names
        }

        # pyre-ignore[6]
        task_id = self.lst_flag_config.index(flag_config)

        active_param_names = self.lst_active_param_names[task_id]

        y = self.lst_xgb_models[task_id].predict(
            np.array([params[name] for name in active_param_names])[None, :]
        )

        return torch.from_numpy(y).reshape(1, 1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any

from ax.generators.torch.botorch_modular.acquisition import Acquisition
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.multioutput_acquisition import (
    MultiOutputAcquisitionFunctionWrapper,
)


class MultiAcquisition(Acquisition):
    """
    A MultiAcquisition class for generating points by optimizing multiple
    acquisition functions jointly.
    """

    def _instantiate_acquisition(
        self,
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ],
    ) -> None:
        """Constructs the acquisition function based on the provided AF clases.

        Args:
            botorch_acqf_classes: A list of BoTorch acquisition function classes.
        """
        self.acq_function_sequence = None
        if len(botorch_acqf_classes_with_options) > 1:
            acqfs = [
                self._construct_botorch_acquisition(
                    botorch_acqf_class=botorch_acqf_class,
                    botorch_acqf_options=botorch_acqf_options,
                    model=self._model,
                )
                for botorch_acqf_class, botorch_acqf_options in (
                    botorch_acqf_classes_with_options
                )
            ]
            self.acqf = MultiOutputAcquisitionFunctionWrapper(acqfs=acqfs)
            self.models_used = [self.surrogate.model_name_by_metric]
        else:
            # Using one acqf with multiple models.
            botorch_acqf_class, botorch_acqf_options = (
                botorch_acqf_classes_with_options[0]
            )
            # Default acqf is the surrogate default.
            self.acqf = self._construct_botorch_acquisition(
                botorch_acqf_class=botorch_acqf_class,
                botorch_acqf_options=botorch_acqf_options,
                model=self._model,
            )
            self.models_used = [self.surrogate.model_name_by_metric]
            if self.n is not None and self.n > 1:
                # Using multiple models
                models_used, models = self.surrogate.models_for_gen(n=self.n)
                # Either 1 or n models will be returned. If 1, we do nothing.
                if len(models) == self.n:
                    self.acq_function_sequence = []
                    for model in models:
                        model, _, _, _ = self._subset_model(
                            model=model, objective_weights=self._full_objective_weights
                        )
                        acqf = self._construct_botorch_acquisition(
                            botorch_acqf_class=botorch_acqf_class,
                            botorch_acqf_options=botorch_acqf_options,
                            model=model,
                        )
                        self.acq_function_sequence.append(acqf)
                    self.models_used = models_used



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.plotly.color_constants import COLOR_FOR_DECREASES, COLOR_FOR_INCREASES
from ax.analysis.plotly.plotly_analysis import create_plotly_analysis_card

from ax.analysis.utils import extract_relevant_adapter
from ax.core.experiment import Experiment
from ax.core.parameter import ChoiceParameter
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.plot.helper import get_plot_data
from ax.utils.stats.statstools import marginal_effects
from plotly import graph_objects as go
from pyre_extensions import none_throws, override


MARGINAL_EFFECTS_CARDGROUP_TITLE = "Marginal Effects For Each Parameter Factor Level"

MARGINAL_EFFECTS_CARDGROUP_SUBTITLE = (
    "This analysis has a bar chart for each parameter, showing the predicted "
    "relative change in the metric for each factor level of the parameter. "
    "The bars are colored based on whether the effect is positive or negative, "
    "and the error bars represent the standard error of the effect."
)


class MarginalEffectsPlot(Analysis):
    """
    Plotly bar charts showing the marginal effect of each factor level of
    selected `ChoiceParameters` on the given metric. This plot is useful
    for understanding how each parameter value marginally affects the metric.

    The DataFrame computed will contain the following columns:
        - Name: The name of the parameter.
        - Level: The level of the parameter.
        - Beta: The effect size, which is the relative (in %) change in the
            metric when the specified factor is used compared to the baseline
            which is marginalized over all factors and levels.
        - SE: The standard error.
    """

    def __init__(
        self,
        metric_name: str,
        parameters: list[str] | None = None,
    ) -> None:
        """
        Args:
            metric_name: The name of the metric to plot.
            parameters: The names of the `ChoiceParameters` to include in the
            analysis. If not specified, all `ChoiceParameters` in the Experiment
            will be included.
        """

        self.metric_name = metric_name
        self.parameters = parameters

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        if experiment is None:
            raise UserInputError("MarginalEffectsPlot requires an Experiment")

        # Either extract ChoiceParameters from experiment or check the ones passed in
        if self.parameters is None:
            self.parameters = [
                param_name
                for param_name, param in experiment.parameters.items()
                if isinstance(param, ChoiceParameter)
            ]
            if len(self.parameters) == 0:
                raise UserInputError(
                    "MarginalEffectsPlot is only for `ChoiceParameter`s, "
                    "but no ChoiceParameters were found in the experiment."
                )
        else:
            for param_name in none_throws(self.parameters):
                parameter = experiment.parameters.get(param_name)
                if not isinstance(parameter, ChoiceParameter):
                    raise UserInputError(
                        "MarginalEffectsPlot is only for `ChoiceParameter`s, but got."
                        f"'{param_name}' which is of type {type(parameter).__name__}."
                    )

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )
        df = _prepare_data(
            metric=self.metric_name,
            adapter=relevant_adapter,
            parameters=self.parameters,
        )
        # Create a card for each parameter
        cards = []
        for param in df["Name"].unique():
            param_df = df[df["Name"] == param]
            fig = _prepare_plot(param_df, self.metric_name)
            cards.append(
                create_plotly_analysis_card(
                    name=self.__class__.__name__,
                    title=f"Marginal Effects for {param}",
                    subtitle=(
                        "This plot visualizes the predicted relative changes in "
                        f"{self.metric_name} for each factor level of {param}, "
                        "offering insights into their impact on the metric. "
                        "By comparing the effects of different levels, this plot "
                        "helps identify which factor levels have the most "
                        "significant influence, providing a detailed understanding "
                        "of the experimental results and guiding future decisions on "
                        "factor selection."
                    ),
                    df=param_df,
                    fig=fig,
                )
            )
        return self._create_analysis_card_group_or_card(
            title=MARGINAL_EFFECTS_CARDGROUP_TITLE,
            subtitle=MARGINAL_EFFECTS_CARDGROUP_SUBTITLE,
            children=cards,
        )


def compute_marginal_effects_adhoc(
    metric_name: str,
    experiment: Experiment,
    parameters: list[str] | None = None,
    generation_strategy: GenerationStrategy | None = None,
    adapter: Adapter | None = None,
) -> AnalysisCardBase:
    """
    Helper method to expose adhoc marginal effects plotting. Only for
    advanced users in a notebook setting.

    Args:
        metric_name: The name of the metric to plot.
        parameters: The names of the `ChoiceParameters` to include in the
        analysis. If not specified, all `ChoiceParameters` in the Experiment
        will be included.
        experiment: The experiment to source data from.
        generation_strategy: Optional. The generation strategy to extract
            the adapter from.
        adapter: Optional. The adapter to use for predictions.
    """

    analysis = MarginalEffectsPlot(metric_name=metric_name, parameters=parameters)

    return analysis.compute(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )


def _prepare_data(
    metric: str,
    adapter: Adapter,
    parameters: list[str] | None = None,
) -> pd.DataFrame:
    """
    Prepare the in sample data for the marginal effects plot, then call
    `marginal_effects()` to create a table of the effects containing each
    factor level for the specified metric. The df returned will have the columns:
        - Name: The name of the parameter.
        - Level: The level of the parameter.
        - Beta: The effect size, which is the relative change in the
            metric when the specified factor is used compared to the baseline
            which is marginalized over all factors and levels.
        - SE: The standard error.
    """

    # TODO (T230319678, enofrey): replace legacy `get_plot_data` method
    plot_data, _, _ = get_plot_data(
        model=adapter, generator_runs_dict={}, metric_names={metric}
    )

    arm_dfs = []
    for arm in plot_data.in_sample.values():
        arm_df = pd.DataFrame(arm.parameters, index=[arm.name])
        arm_df["mean"] = arm.y_hat[metric]
        arm_df["sem"] = arm.se_hat[metric]
        arm_dfs.append(arm_df)
    effect_table = marginal_effects(
        df=pd.concat(arm_dfs, axis=0), covariates=parameters
    )
    if parameters is not None:
        effect_table = effect_table[effect_table["Name"].isin(parameters)]

    return effect_table


def _prepare_plot(df: pd.DataFrame, metric_name: str) -> go.Figure:
    """Create a bar plot of the marginal effects for a single parameter."""
    param = df["Name"].iloc[0]
    colors = [
        COLOR_FOR_INCREASES if beta > 0 else COLOR_FOR_DECREASES for beta in df["Beta"]
    ]

    bar_plot = go.Bar(
        x=df["Level"],
        y=df["Beta"]
        / 100,  # Divide by 100 (multiplied 100x later in percentage formatting)
        error_y={
            "type": "data",
            "array": df["SE"] / 100,
        },
        marker={"color": colors},
    )

    fig = go.Figure(data=[bar_plot])
    fig.update_layout(
        title=f"Marginal Effects for {param} on {metric_name}",
        xaxis_title="Factor Level",
        yaxis_title="Effect relative to experiment average",
        yaxis_tickformat=".2%",
        yaxis_hoverformat=".2%",
        showlegend=False,
    )
    return fig



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

import numpy as np
from ax.adapter.base import Adapter, DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.trial_as_task import TrialAsTask
from ax.core.arm import Arm
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.exceptions.core import UnsupportedError
from ax.generators.base import Generator
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment, get_robust_search_space
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


class TrialAsTaskTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.exp = get_branin_experiment(with_status_quo=True, with_batch=True)
        self.adapter = Adapter(
            search_space=self.exp.search_space,
            generator=Generator(),
            experiment=self.exp,
        )
        self.exp.new_batch_trial().add_arm(
            Arm(parameters={"x1": 0, "x2": 0}, name="status_quo")
        ).add_arm(Arm(parameters={"x1": 1, "x2": 1}))
        self.exp.new_batch_trial().add_arm(
            Arm(parameters={"x1": 0, "x2": 0}, name="status_quo")
        ).add_arm(Arm(parameters={"x1": 3, "x2": 3}))
        for t in self.exp.trials.values():
            t.mark_running(no_runner_required=True)
        self.exp.trials[0].mark_completed()
        self.exp.fetch_data()

        self.experiment_data = extract_experiment_data(
            experiment=self.exp, data_loader_config=DataLoaderConfig()
        )
        self.training_feats = [
            ObservationFeatures({"x1": 1, "x2": 1}, trial_index=0),
            ObservationFeatures({"x1": 2, "x2": 2}, trial_index=0),
            ObservationFeatures({"x1": 3, "x2": 3}, trial_index=1),
            ObservationFeatures({"x1": 4, "x2": 4}, trial_index=2),
        ]
        self.training_obs = [
            Observation(
                data=ObservationData(
                    metric_names=[], means=np.array([]), covariance=np.empty((0, 0))
                ),
                features=obsf,
            )
            for obsf in self.training_feats
        ]

        self.t = TrialAsTask(
            search_space=self.exp.search_space,
            experiment_data=self.experiment_data,
            adapter=self.adapter,
        )
        self.bm = {
            "bp1": {0: "v1", 1: "v2", 2: "v3"},
            "bp2": {0: "u1", 1: "u1", 2: "u2"},
        }

        self.t2 = TrialAsTask(
            search_space=self.exp.search_space,
            observations=self.training_obs,
            adapter=self.adapter,
            config={"trial_level_map": self.bm},
        )
        self.t3 = TrialAsTask(
            search_space=self.exp.search_space,
            observations=self.training_obs,
            adapter=self.adapter,
            config={"trial_level_map": {}},
        )
        # test string trial indices
        self.bm2 = {
            p_name: {str(k): v for k, v in value_dict.items()}
            for p_name, value_dict in self.bm.items()
        }
        self.t4 = TrialAsTask(
            search_space=self.exp.search_space,
            observations=self.training_obs,
            adapter=self.adapter,
            config={"trial_level_map": self.bm2, "target_trial": 2},
        )

    def test_Init(self) -> None:
        self.assertEqual(
            self.t.trial_level_map, {"TRIAL_PARAM": {i: str(i) for i in range(3)}}
        )
        self.assertEqual(self.t.inverse_map, {str(i): i for i in range(3)})
        # based on `get_target_trial_index``, the longest running trial is trial 1
        self.assertEqual(self.t.target_values, {"TRIAL_PARAM": "1"})
        self.assertEqual(self.t2.trial_level_map, self.bm)
        self.assertIsNone(self.t2.inverse_map)
        self.assertEqual(self.t2.target_values, {"bp1": "v2", "bp2": "u1"})
        # check that strings were converted to integers
        self.assertEqual(self.t4.trial_level_map, self.bm)
        self.assertIsNone(self.t4.inverse_map)
        self.assertEqual(self.t4.target_values, {"bp1": "v3", "bp2": "u2"})
        # Test validation
        obsf = ObservationFeatures({"x1": 2, "x2": 2})
        obs = Observation(
            data=ObservationData([], np.array([]), np.empty((0, 0))), features=obsf
        )
        with self.assertRaises(ValueError):
            TrialAsTask(
                search_space=self.exp.search_space,
                observations=self.training_obs + [obs],
                adapter=self.adapter,
            )
        bm = {"p": {0: "y", 1: "z"}}
        with self.assertRaises(ValueError):
            TrialAsTask(
                search_space=self.exp.search_space,
                observations=self.training_obs,
                adapter=self.adapter,
                config={"trial_level_map": bm},
            )

    def test_TransformObservationFeatures(self) -> None:
        obs_ft1 = deepcopy(self.training_feats)
        obs_ft2 = deepcopy(self.training_feats)
        obs_ft_trans1 = [
            ObservationFeatures({"x1": 1, "x2": 1, "TRIAL_PARAM": "0"}),
            ObservationFeatures({"x1": 2, "x2": 2, "TRIAL_PARAM": "0"}),
            ObservationFeatures({"x1": 3, "x2": 3, "TRIAL_PARAM": "1"}),
            ObservationFeatures({"x1": 4, "x2": 4, "TRIAL_PARAM": "2"}),
        ]
        obs_ft_trans2 = [
            ObservationFeatures({"x1": 1, "x2": 1, "bp1": "v1", "bp2": "u1"}),
            ObservationFeatures({"x1": 2, "x2": 2, "bp1": "v1", "bp2": "u1"}),
            ObservationFeatures({"x1": 3, "x2": 3, "bp1": "v2", "bp2": "u1"}),
            ObservationFeatures({"x1": 4, "x2": 4, "bp1": "v3", "bp2": "u2"}),
        ]
        obs_ft1 = self.t.transform_observation_features(obs_ft1)
        self.assertEqual(obs_ft1, obs_ft_trans1)
        obs_ft2 = self.t2.transform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, obs_ft_trans2)
        obs_ft1 = self.t.untransform_observation_features(obs_ft1)
        self.assertEqual(obs_ft1, self.training_feats)
        obs_ft3 = self.t2.transform_observation_features([ObservationFeatures({})])
        self.assertEqual(obs_ft3[0], ObservationFeatures({}))
        obs_ft4 = deepcopy(self.training_feats)
        obs_ft4 = self.t3.untransform_observation_features(obs_ft4)
        self.assertEqual(obs_ft4, self.training_feats)

    def test_TransformObservationFeaturesWithoutTrialIndex(self) -> None:
        obs_ft_no_trial_index = deepcopy(self.training_feats)
        obs_ft_no_trial_index.append(
            ObservationFeatures(
                {"x1": 20, "x2": 20},
            )
        )
        obs_ft_trans = [
            ObservationFeatures({"x1": 1, "x2": 1, "TRIAL_PARAM": "0"}),
            ObservationFeatures({"x1": 2, "x2": 2, "TRIAL_PARAM": "0"}),
            ObservationFeatures({"x1": 3, "x2": 3, "TRIAL_PARAM": "1"}),
            ObservationFeatures({"x1": 4, "x2": 4, "TRIAL_PARAM": "2"}),
            ObservationFeatures({"x1": 20, "x2": 20, "TRIAL_PARAM": "2"}),
        ]
        obs_ft_trans2 = [
            ObservationFeatures({"x1": 1, "x2": 1, "bp1": "v1", "bp2": "u1"}),
            ObservationFeatures({"x1": 2, "x2": 2, "bp1": "v1", "bp2": "u1"}),
            ObservationFeatures({"x1": 3, "x2": 3, "bp1": "v2", "bp2": "u1"}),
            ObservationFeatures({"x1": 4, "x2": 4, "bp1": "v3", "bp2": "u2"}),
            ObservationFeatures({"x1": 20, "x2": 20, "bp1": "v3", "bp2": "u2"}),
        ]

        # test can transform and untransform with no config
        obs_ft_no_trial_index_transformed = self.t.transform_observation_features(
            obs_ft_no_trial_index
        )
        self.assertEqual(obs_ft_no_trial_index_transformed, obs_ft_trans)
        untransformed = self.t.untransform_observation_features(
            obs_ft_no_trial_index_transformed
        )
        # test can transform and untransform with config trial level map
        self.assertEqual(untransformed, obs_ft_no_trial_index)
        obs_ft_no_index_transformed_2 = self.t2.transform_observation_features(
            obs_ft_no_trial_index
        )
        self.assertEqual(obs_ft_no_index_transformed_2, obs_ft_trans2)
        # can transform and untransform are equal with empty config
        obs_ft4 = self.t3.untransform_observation_features(obs_ft_no_trial_index)
        self.assertEqual(obs_ft4, obs_ft_no_trial_index)

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.exp.search_space)
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(set(ss2.parameters.keys()), {"x1", "x2", "TRIAL_PARAM"})
        p = assert_is_instance(ss2.parameters["TRIAL_PARAM"], ChoiceParameter)
        self.assertEqual(p.parameter_type, ParameterType.STRING)
        self.assertEqual(set(p.values), {"0", "1", "2"})
        self.assertTrue(p.is_task)
        self.assertFalse(p.is_ordered)
        self.assertEqual(p.target_value, "1")
        ss2 = deepcopy(self.exp.search_space)
        ss2 = self.t2.transform_search_space(ss2)
        self.assertEqual(set(ss2.parameters.keys()), {"x1", "x2", "bp1", "bp2"})
        p = ss2.parameters["bp1"]
        self.assertTrue(isinstance(p, ChoiceParameter))
        self.assertEqual(p.parameter_type, ParameterType.STRING)
        self.assertEqual(set(p.values), {"v1", "v2", "v3"})
        self.assertTrue(p.is_task)
        self.assertFalse(p.is_ordered)
        self.assertEqual(p.target_value, "v2")
        p = ss2.parameters["bp2"]
        self.assertTrue(isinstance(p, ChoiceParameter))
        self.assertEqual(p.parameter_type, ParameterType.STRING)
        self.assertEqual(set(p.values), {"u1", "u2"})
        self.assertTrue(p.is_task)
        self.assertTrue(p.is_ordered)  # 2 choices so always ordered
        self.assertEqual(p.target_value, "u1")
        t = TrialAsTask(
            search_space=self.exp.search_space,
            observations=self.training_obs,
            adapter=self.adapter,
            config={
                "trial_level_map": {
                    "trial_index": {0: 10, 1: 11, 2: 12},
                },
                "target_trial": 1,
            },
        )
        ss2 = deepcopy(self.exp.search_space)
        ss2 = t.transform_search_space(ss2)
        p = assert_is_instance(ss2.parameters["trial_index"], ChoiceParameter)
        self.assertEqual(p.parameter_type, ParameterType.INT)
        self.assertEqual(set(p.values), {10, 11, 12})
        self.assertTrue(p.is_ordered)
        self.assertTrue(p.is_task)
        self.assertEqual(p.target_value, 11)

    def test_w_robust_search_space(self) -> None:
        rss = get_robust_search_space()
        # Raises an error in __init__.
        with self.assertRaisesRegex(UnsupportedError, "transform is not supported"):
            TrialAsTask(
                search_space=rss,
                observations=self.training_obs,
                adapter=self.adapter,
            )

    def test_less_than_two_trials(self) -> None:
        # test transform is a no-op with less than two trials
        exp = get_branin_experiment()
        exp.new_trial().add_arm(Arm(parameters={"x1": 1, "x2": 1}))
        adapter = Adapter(
            search_space=exp.search_space,
            generator=Generator(),
            experiment=exp,
        )
        training_obs = self.training_obs[:1]
        t = TrialAsTask(
            search_space=exp.search_space,
            observations=training_obs,
            adapter=adapter,
        )
        self.assertEqual(t.trial_level_map, {})
        training_feats = [training_obs[0].features]
        training_feats_clone = deepcopy(training_feats)
        self.assertEqual(
            t.transform_observation_features(training_feats_clone), training_feats
        )
        self.assertEqual(
            t.untransform_observation_features(training_feats), training_feats_clone
        )
        ss2 = exp.search_space.clone()
        self.assertEqual(t.transform_search_space(ss2), exp.search_space)

    def test_less_than_two_levels(self) -> None:
        # test transform is a no-op with less than two trials
        exp = get_branin_experiment()
        exp.new_trial().add_arm(Arm(parameters={"x1": 1, "x2": 1}))
        exp.new_trial().add_arm(Arm(parameters={"x1": 2, "x2": 2}))
        adapter = Adapter(
            search_space=exp.search_space,
            generator=Generator(),
            experiment=exp,
        )
        training_obs = self.training_obs[:1]
        t = TrialAsTask(
            search_space=exp.search_space,
            observations=training_obs,
            adapter=adapter,
            config={"trial_level_map": {"t": {0: "v1", 1: "v1"}}},
        )
        self.assertEqual(t.trial_level_map, {})
        training_feats = [training_obs[0].features]
        training_feats_clone = deepcopy(training_feats)
        self.assertEqual(
            t.transform_observation_features(training_feats_clone), training_feats
        )
        self.assertEqual(
            t.untransform_observation_features(training_feats), training_feats_clone
        )
        ss2 = exp.search_space.clone()
        self.assertEqual(t.transform_search_space(ss2), exp.search_space)

    def test_transform_experiment_data(self) -> None:
        # Experiment data has 16 rows for trial 0 and 2 rows each for trials 1 & 2.
        def make_expected(values: tuple[str, str, str]) -> list[str]:
            return [values[0]] * 16 + [values[1]] * 2 + [values[2]] * 2

        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Check that arm data has trial parameter added.
        self.assertEqual(
            transformed_data.arm_data["TRIAL_PARAM"].to_list(),
            make_expected(values=("0", "1", "2")),
        )
        # Check that other columns are unchanged.
        assert_frame_equal(
            transformed_data.arm_data.drop(columns="TRIAL_PARAM"),
            self.experiment_data.arm_data,
        )
        # Check that observation data is unchanged.
        assert_frame_equal(
            transformed_data.observation_data, self.experiment_data.observation_data
        )

        # Test with alternative transform config.
        transformed_data = self.t2.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        self.assertEqual(
            transformed_data.arm_data["bp1"].to_list(),
            make_expected(values=("v1", "v2", "v3")),
        )
        self.assertEqual(
            transformed_data.arm_data["bp2"].to_list(),
            make_expected(values=("u1", "u1", "u2")),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING

import pandas as pd
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.generators.types import TConfig
from ax.utils.common.timeutils import unixtime_to_pandas_ts
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class TimeAsFeature(Transform):
    """Convert start time and duration into features that can be used for modeling.

    If no end_time is available, the current time is used.

    Duration is normalized to the unit cube.

    Transform is done in-place.

    TODO: revise this when better support for non-tunable features is added.
    """

    requires_data_for_initialization: bool = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        if isinstance(search_space, RobustSearchSpace):
            raise UnsupportedError(
                "TimeAsFeature transform is not supported for RobustSearchSpace."
            )
        self.min_start_time: float = float("inf")
        self.max_start_time: float = float("-inf")
        self.min_duration: float = float("inf")
        self.max_duration: float = float("-inf")
        self.current_time: float = time()
        if observations is not None:
            for obs in observations:
                obsf = obs.features
                if obsf.start_time is None:
                    raise ValueError(
                        "Unable to use TimeAsFeature since not all observations have "
                        "start time specified."
                    )
                start_time = none_throws(obsf.start_time).timestamp()
                self.min_start_time = min(self.min_start_time, start_time)
                self.max_start_time = max(self.max_start_time, start_time)
                duration = self._get_duration(
                    start_time=start_time, end_time=obsf.end_time
                )
                self.min_duration = min(self.min_duration, duration)
                self.max_duration = max(self.max_duration, duration)
        else:
            obs_data = none_throws(experiment_data).observation_data
            if len(obs_data.index.names) > 2:
                raise UnsupportedError(
                    "TimeAsFeature transform is not supported with map data."
                )
            # Dividing by 1e9 to convert from nanoseconds to seconds, to be consistent
            # with usage of seconds in the `Observation` version.
            start_times = obs_data[("metadata", "start_time")].astype("int64") / 1e9
            if start_times.isna().any():
                raise ValueError(
                    "Unable to use TimeAsFeature since not all observations have "
                    "start time specified."
                )
            current_time_ts = unixtime_to_pandas_ts(self.current_time)
            end_times = (
                obs_data[("metadata", "end_time")]
                .fillna(current_time_ts)
                .astype("int64")
                / 1e9
            )
            durations = end_times - start_times
            self.min_start_time = start_times.min()
            self.max_start_time = start_times.max()
            self.min_duration = durations.min()
            self.max_duration = durations.max()
        self.duration_range: float = self.max_duration - self.min_duration
        if self.duration_range == 0:
            # Avoid division by zero issues when the duration is constant.
            self.duration_range = 1.0

    def _get_duration(self, start_time: float, end_time: pd.Timestamp | None) -> float:
        return (
            self.current_time if end_time is None else end_time.timestamp()
        ) - start_time

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            if obsf.start_time is not None:
                start_time = obsf.start_time.timestamp()
                obsf.parameters["start_time"] = start_time
                duration = self._get_duration(
                    start_time=start_time, end_time=obsf.end_time
                )
                # normalize duration to the unit cube
                obsf.parameters["duration"] = (
                    duration - self.min_duration
                ) / self.duration_range
            else:
                # start time can be None for pending arms that generated
                # with a model that did not use the TimeAsFeature transform.
                # In that case, assume the arm is going to be evaluated at the
                # current time, and that the duration is the midpoint of the
                # range.
                obsf.parameters["start_time"] = self.current_time
                obsf.parameters["duration"] = 0.5
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for p_name in ("start_time", "duration"):
            if p_name in search_space.parameters:
                raise ValueError(
                    f"Parameter name {p_name} is reserved when using "
                    "TimeAsFeature transform, but is part of the provided "
                    "search space. Please choose a different name for "
                    "this parameter."
                )
        param = RangeParameter(
            name="start_time",
            parameter_type=ParameterType.FLOAT,
            lower=self.min_start_time,
            upper=self.max_start_time,
        )
        search_space.add_parameter(param)
        param = RangeParameter(
            name="duration",
            parameter_type=ParameterType.FLOAT,
            # duration is normalized to [0,1]
            lower=0.0,
            upper=1.0,
        )
        search_space.add_parameter(param)
        return search_space

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            start_time = obsf.parameters.pop("start_time", None)
            duration = obsf.parameters.pop("duration", None)
            if start_time is not None:
                start_time = assert_is_instance(start_time, float)
                obsf.start_time = unixtime_to_pandas_ts(start_time)
                if duration is not None:
                    duration = assert_is_instance(duration, float)
                    obsf.end_time = unixtime_to_pandas_ts(
                        duration * self.duration_range + self.min_duration + start_time
                    )
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        obs_data = experiment_data.observation_data
        # Dividing by 1e9 to convert from nanoseconds to seconds.
        start_times = obs_data[("metadata", "start_time")].astype("int64") / 1e9
        current_time_ts = unixtime_to_pandas_ts(self.current_time)
        end_times = (
            obs_data[("metadata", "end_time")].fillna(current_time_ts).astype("int64")
            / 1e9
        )
        experiment_data.arm_data["start_time"] = start_times
        duration = end_times - start_times
        duration = (duration - self.min_duration) / self.duration_range
        experiment_data.arm_data["duration"] = duration
        return experiment_data



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import markdown
import pandas as pd

from ax.analysis.analysis_card import AnalysisCard
from IPython.display import Markdown


class MarkdownAnalysisCard(AnalysisCard):
    def get_markdown(self) -> str:
        return self.blob

    def _body_html(self, depth: int) -> str:
        return f"<div class='content'>{markdown.markdown(self.get_markdown())}<div>"

    def _body_papermill(self) -> Markdown:
        return Markdown(self.get_markdown())


def create_markdown_analysis_card(
    name: str,
    title: str,
    subtitle: str,
    df: pd.DataFrame,
    message: str,
) -> MarkdownAnalysisCard:
    return MarkdownAnalysisCard(
        name=name,
        title=title,
        subtitle=subtitle,
        df=df,
        blob=message,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterable
from datetime import datetime, timedelta
from random import randint
from typing import Any

from ax.core.base_trial import BaseTrial, TrialStatus
from ax.runners.single_running_trial_mixin import SingleRunningTrialMixin
from ax.runners.synthetic import SyntheticRunner
from ax.service.orchestrator import Orchestrator
from pyre_extensions import none_throws

DUMMY_EXCEPTION = "test_exception"
TEST_MEAN = 1.0


class SyntheticRunnerWithStatusPolling(SyntheticRunner):
    """Test runner that implements `poll_trial_status`, required for compatibility
    with the ``Orchestrator``."""

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        # Pretend that sometimes trials take a few seconds to complete and that they
        # might get completed out of order.
        if randint(0, 3) > 0:
            running = [t.index for t in trials]
            return {TrialStatus.COMPLETED: {running[randint(0, len(running) - 1)]}}
        return {}


class SyntheticRunnerWithSingleRunningTrial(SingleRunningTrialMixin, SyntheticRunner):
    pass


class SyntheticRunnerWithPredictableStatusPolling(SyntheticRunner):
    """Test runner that implements `poll_trial_status`, required for compatibility
    with the ``Orchestrator``, which polls completed."""

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        completed = {t.index for t in trials}
        return {TrialStatus.COMPLETED: completed}


class TestOrchestrator(Orchestrator):
    """Test Orchestrator that only implements ``report_results`` for convenience in
    testing.
    """

    def report_results(self, force_refit: bool = False) -> dict[str, Any]:
        return {
            # Use `set` constructor to copy the set, else the value
            # will be a pointer and all will be the same.
            "trials_completed_so_far": set(
                self.experiment.trial_indices_by_status[TrialStatus.COMPLETED]
            ),
            "trials_early_stopped_so_far": set(
                self.experiment.trial_indices_by_status[TrialStatus.EARLY_STOPPED]
            ),
        }


# ---- Runners below simulate different usage and failure modes for Orchestrator ----


class RunnerWithFrequentFailedTrials(SyntheticRunner):
    poll_failed_next_time = True

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        running = [t.index for t in trials]
        status = (
            TrialStatus.FAILED if self.poll_failed_next_time else TrialStatus.COMPLETED
        )
        # Poll different status next time.
        self.poll_failed_next_time = not self.poll_failed_next_time
        return {status: {running[randint(0, len(running) - 1)]}}


class RunnerWithFailedAndAbandonedTrials(SyntheticRunner):
    poll_failed_next_time = True
    status_idx = 0

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        trial_statuses_dummy = {
            0: TrialStatus.ABANDONED,
            1: TrialStatus.FAILED,
            2: TrialStatus.COMPLETED,
        }
        running = [t.index for t in trials]
        status = trial_statuses_dummy[self.status_idx]

        # Poll different status next time.
        self.status_idx = (self.status_idx + 1) % 3

        return {status: {running[randint(0, len(running) - 1)]}}


class RunnerWithAllFailedTrials(SyntheticRunner):
    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        running = [t.index for t in trials]
        return {TrialStatus.FAILED: {running[randint(0, len(running) - 1)]}}


class NoReportResultsRunner(SyntheticRunner):
    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        if randint(0, 3) > 0:
            running = [t.index for t in trials]
            return {TrialStatus.COMPLETED: {running[randint(0, len(running) - 1)]}}
        return {}


class InfinitePollRunner(SyntheticRunner):
    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        return {}


class RunnerWithEarlyStoppingStrategy(SyntheticRunner):
    poll_trial_status_count = 0

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        self.poll_trial_status_count += 1

        # In the first step, don't complete any trials
        # Trial #1 will be early stopped
        if self.poll_trial_status_count == 1:
            return {}

        if self.poll_trial_status_count == 2:
            return {TrialStatus.COMPLETED: {2}}

        return {TrialStatus.COMPLETED: {0}}


class BrokenRunnerValueError(SyntheticRunnerWithStatusPolling):
    run_trial_call_count = 0

    def run_multiple(self, trials: Iterable[BaseTrial]) -> dict[int, dict[str, Any]]:
        self.run_trial_call_count += 1
        raise ValueError("Failing for testing purposes.")


class BrokenRunnerRuntimeError(SyntheticRunnerWithStatusPolling):
    run_trial_call_count = 0

    def run_multiple(self, trials: Iterable[BaseTrial]) -> dict[int, dict[str, Any]]:
        self.run_trial_call_count += 1
        raise RuntimeError("Failing for testing purposes.")


class RunnerToAllowMultipleMapMetricFetches(SyntheticRunnerWithStatusPolling):
    """``Runner`` that gives a trial 3 seconds to run before considering
    the trial completed, which gives us some time to fetch the ``MapMetric``
    a few times, if there is one on the experiment. Useful for testing behavior
    with repeated ``MapMetric`` fetches.
    """

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        running_trials = next(iter(trials)).experiment.trials_by_status[
            TrialStatus.RUNNING
        ]
        completed, still_running = set(), set()
        for t in running_trials:
            if datetime.now() - none_throws(t.time_run_started) > timedelta(seconds=3):
                completed.add(t.index)
            else:
                still_running.add(t.index)

        return {
            TrialStatus.COMPLETED: completed,
            TrialStatus.RUNNING: still_running,
        }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings
from math import isnan
from typing import Any, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.metadata_to_float import MetadataToFloat
from ax.core.map_data import MAP_KEY
from ax.core.observation import Observation, ObservationFeatures
from ax.core.search_space import SearchSpace
from ax.core.utils import has_map_metrics
from ax.generators.types import TConfig
from pandas import Index, MultiIndex
from pyre_extensions import none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class MapKeyToFloat(MetadataToFloat):
    """
    This transform extracts the entry from the metadata field of the observation
    features corresponding to the `parameters` specified in the transform config
    and inserts it into the parameter field. If no parameters are specified in the
    config, the transform will extract the map key name from the optimization config.

    Inheriting from the `MetadataToFloat` transform, this transform also adds a range
    (float) parameter to the search space. Similarly, users can override the default
    behavior by specifying the `config` with `parameters` as the key, where each entry
    maps a metadata key to a dictionary of keyword arguments for the corresponding
    RangeParameter constructor. NOTE: log and logit-scale options are not supported.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        config = config or {}
        if "parameters" not in config:
            # Extract map keys from the optimization config, if no parameters are
            # specified in the config.
            if (
                adapter is not None
                and adapter._optimization_config is not None
                and has_map_metrics(optimization_config=adapter._optimization_config)
            ):
                config["parameters"] = {MAP_KEY: {}}
            else:
                warnings.warn(
                    (
                        f"{self.__class__.__name__} is unable to identify `parameters` "
                        "in the transform config or an adapter with an "
                        "optimization config from which the map keys can be inferred; "
                        "this transform will not perform any operations and behave as "
                        "a no-op."
                    ),
                    stacklevel=2,
                )
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )

    def _get_values_for_parameter(
        self,
        name: str,
        observations: list[Observation] | None,
        experiment_data: ExperimentData | None,
    ) -> set[float]:
        if experiment_data is not None:
            obs_data = experiment_data.observation_data
            if name not in obs_data.index.names:
                raise ValueError(
                    f"Parameter {name} is not in the index of the observation data."
                )
            return set(
                obs_data.index.unique(level=name).dropna().astype(float).tolist()
            )
        # For Observations, the logic is identical to the parent class.
        return super()._get_values_for_parameter(
            name=name,
            observations=observations,
            experiment_data=experiment_data,
        )

    def _transform_observation_feature(self, obsf: ObservationFeatures) -> None:
        if len(obsf.parameters) == 0:
            obsf.parameters = {p.name: p.upper for p in self._parameter_list}
            return
        if obsf.metadata is None:
            obsf.metadata = {}
        metadata: dict[str, Any] = none_throws(obsf.metadata)
        for p in self._parameter_list:
            if isnan(metadata.get(p.name, float("nan"))):
                metadata[p.name] = p.upper
        super()._transform_observation_feature(obsf)

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        """Transform experiment data by replacing NaN/null values
        in the corresponding index of observation_data with the upper bound
        of the parameter.

        This reproduces the behavior of `_transform_observation_feature` when
        the parameter is missing in the observation features.

        In contrast with other transforms that extract features from various
        metadata, this transform does not add a new column to arm_data. Instead,
        the parameter values are directly extracted from the index of the
        observation data, as long as the parameter is added to the search space.
        """
        if not self._parameter_list:
            return experiment_data

        observation_data = experiment_data.observation_data

        # Create a mapping of parameter names to their upper bounds for quick lookup
        param_upper_bounds = {p.name: p.upper for p in self._parameter_list}

        # Build new index arrays, filling NaN values where needed.
        new_levels: list[Index] = []
        index_needs_update = False

        for level_name in observation_data.index.names:
            level_values = observation_data.index.get_level_values(level_name)

            if level_name in param_upper_bounds and level_values.isna().any():
                # Replace NaN/null values with the parameter's upper bound.
                filled_values = level_values.fillna(param_upper_bounds[level_name])
                new_levels.append(filled_values)
                index_needs_update = True
            else:
                # Keep original level values.
                new_levels.append(level_values)

        # Update the index only if any levels needed updating.
        if index_needs_update:
            observation_data.index = MultiIndex.from_arrays(
                new_levels, names=observation_data.index.names
            )

        return ExperimentData(
            arm_data=experiment_data.arm_data,
            observation_data=observation_data,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import torch
from ax.adapter.registry import Generators
from ax.api.utils.structs import GenerationStrategyDispatchStruct
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import UnsupportedError
from ax.generation_strategy.center_generation_node import CenterGenerationNode
from ax.generation_strategy.generation_strategy import (
    GenerationNode,
    GenerationStrategy,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import MinTrials
from ax.generators.torch.botorch_modular.surrogate import ModelConfig, SurrogateSpec
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP


def _get_sobol_node(
    initialization_budget: int | None,
    min_observed_initialization_trials: int | None,
    initialize_with_center: bool,
    use_existing_trials_for_initialization: bool,
    allow_exceeding_initialization_budget: bool,
    initialization_random_seed: int | None,
) -> GenerationNode:
    """Constructs a Sobol node based on inputs from
    ``struct``. The Sobol generator utilizes
    `initialization_random_seed` if specified.

    This node always transitions to "MBM", using the following transition criteria:
    - MinTrials enforcing the initialization budget.
        - If the initialization budget is not specified, it defaults to 5.
        - The TC will not block generation if `allow_exceeding_initialization_budget`
            is set to True.
        - The TC is currently not restricted to any trial statuses and will
            count all trials.
        - `use_existing_trials_for_initialization` controls whether trials previously
            attached to the experiment are counted as part of the initialization budget.
    - MinTrials enforcing the minimum number of observed initialization trials.
        - If `min_observed_initialization_trials` is not specified, it defaults
            to `max(1, initialization_budget // 2)`.
        - The TC currently only counts trials in status COMPLETED (with data attached)
            as observed trials.
    """
    # Set the default options.
    initialization_budget = initialization_budget
    if initialization_budget is None:
        initialization_budget = 5
    if min_observed_initialization_trials is None:
        min_observed_initialization_trials = max(1, initialization_budget // 2)
    if initialize_with_center and not use_existing_trials_for_initialization:
        # Account for center point in initialization, since the TC will not count it.
        initialization_budget -= 1
    # Construct the transition criteria.
    transition_criteria = [
        MinTrials(  # This represents the initialization budget.
            threshold=initialization_budget,
            transition_to="MBM",
            block_gen_if_met=(not allow_exceeding_initialization_budget),
            block_transition_if_unmet=True,
            use_all_trials_in_exp=use_existing_trials_for_initialization,
        ),
        MinTrials(  # This represents minimum observed trials requirement.
            threshold=min_observed_initialization_trials,
            transition_to="MBM",
            block_gen_if_met=False,
            block_transition_if_unmet=True,
            use_all_trials_in_exp=True,
            only_in_statuses=[TrialStatus.COMPLETED],
            count_only_trials_with_data=True,
        ),
    ]
    return GenerationNode(
        node_name="Sobol",
        generator_specs=[
            GeneratorSpec(
                generator_enum=Generators.SOBOL,
                model_kwargs={"seed": initialization_random_seed},
            )
        ],
        transition_criteria=transition_criteria,
        should_deduplicate=True,
    )


def _get_mbm_node(
    method: str,
    torch_device: str | None,
) -> GenerationNode:
    """Constructs an MBM node based on the method specified in
    ``struct``.

    The ``SurrogateSpec`` takes the following form for the given method:
    - BALANCED: Two model configs: one with MBM defaults, the other with
        linear kernel with input warping.
    - FAST: An empty model config that utilizes MBM defaults.
    """
    # Construct the surrogate spec.
    if method == "quality":
        model_configs = [
            ModelConfig(
                botorch_model_class=SaasFullyBayesianSingleTaskGP,
                model_options={"use_input_warping": True},
                mll_options={
                    "disable_progbar": True,
                },
                name="WarpedSAAS",
            )
        ]
    elif method == "fast":
        model_configs = [ModelConfig(name="MBM defaults")]
    else:
        raise UnsupportedError(f"Unsupported generation method: {method}.")

    device = None if torch_device is None else torch.device(torch_device)

    return GenerationNode(
        node_name="MBM",
        generator_specs=[
            GeneratorSpec(
                generator_enum=Generators.BOTORCH_MODULAR,
                model_kwargs={
                    "surrogate_spec": SurrogateSpec(model_configs=model_configs),
                    "torch_device": device,
                },
            )
        ],
        should_deduplicate=True,
    )


def choose_generation_strategy(
    struct: GenerationStrategyDispatchStruct,
) -> GenerationStrategy:
    """
    Choose a generation strategy based on the properties of the experiment and the
    inputs provided in ``struct``.

    NOTE: The behavior of this function is subject to change. It will be updated to
    produce best general purpose generation strategies based on benchmarking results.

    Args:
        struct: A ``GenerationStrategyDispatchStruct``
            object that informs
            the choice of generation strategy.

    Returns:
        A generation strategy.
    """
    # Handle the random search case.
    if struct.method == "random_search":
        nodes = [
            GenerationNode(
                node_name="Sobol",
                generator_specs=[
                    GeneratorSpec(
                        generator_enum=Generators.SOBOL,
                        model_kwargs={"seed": struct.initialization_random_seed},
                    )
                ],
            )
        ]
        gs_name = "QuasiRandomSearch"
    else:
        mbm_node = _get_mbm_node(
            method=struct.method,
            torch_device=struct.torch_device,
        )
        if (
            struct.initialization_budget is None
            or struct.initialization_budget > struct.initialize_with_center
        ):
            nodes = [
                _get_sobol_node(
                    initialization_budget=struct.initialization_budget,
                    min_observed_initialization_trials=struct.min_observed_initialization_trials,  # noqa: E501
                    initialize_with_center=struct.initialize_with_center,
                    use_existing_trials_for_initialization=struct.use_existing_trials_for_initialization,  # noqa: E501
                    allow_exceeding_initialization_budget=struct.allow_exceeding_initialization_budget,  # noqa: E501
                    initialization_random_seed=struct.initialization_random_seed,
                ),
                mbm_node,
            ]
            gs_name = f"Sobol+MBM:{struct.method}"
        else:
            nodes = [mbm_node]
            gs_name = f"MBM:{struct.method}"
    if struct.initialize_with_center and (
        struct.initialization_budget is None or struct.initialization_budget > 0
    ):
        center_node = CenterGenerationNode(next_node_name=nodes[0].node_name)
        nodes.insert(0, center_node)
        gs_name = f"Center+{gs_name}"
    return GenerationStrategy(name=gs_name, nodes=nodes)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from enum import Enum, unique
from typing import TYPE_CHECKING

from ax.core.data import Data
from ax.utils.common.base import SortableBase


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import core  # noqa F401


class AuxiliaryExperiment(SortableBase):
    """Class for defining an auxiliary experiment."""

    def __init__(
        self,
        experiment: core.experiment.Experiment,
        is_active: bool = True,
        data: Data | None = None,
    ) -> None:
        """
        Lightweight container of an experiment, and its data,
        that will be used as auxiliary information for another experiment.
        Args:
            experiment: The Ax experiment with the auxiliary data.
            is_active: Whether the auxiliary experiment relation to the target
                experiment is currently active.
            data: Auxiliary data.
        """
        self.experiment = experiment
        self.data: Data = data or experiment.lookup_data()
        self.is_active = is_active

    def _unique_id(self) -> str:
        # While there can be multiple `AuxiliarySource`-s made from the same
        # experiment (and thus sharing the experiment name), the uniqueness
        # here is only needed w.r.t. parent object ("main experiment", for which
        # this will be an auxiliary source for).
        return self.experiment.name


@unique
class AuxiliaryExperimentPurpose(Enum):
    # BOPE Aux Experiment Usage pattern:
    # 1. Run the exploratory batch for the main / BO experiment.
    # 2. Use the BO experiment as the auxiliary experiment for the PE experiment
    #    to construct the outcome model to estimate metric movements during
    #    preference exploration.
    # 3. Use the PE experiment as the auxiliary experiment to get the learned
    #    objective continue iterating with the BO experiment.
    # [BOPE] The BO experiment for which preference exploration is being done.
    BO_EXPERIMENT = "bo_experiment"
    # [BOPE] The preference exploration experiment itself.
    PE_EXPERIMENT = "pe_experiment"
    # [BOTL] The transferable/source experiment for transfer learning.
    TRANSFERABLE_EXPERIMENT = "transferable_experiment"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


# pyre-strict

import pandas as pd

from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckStatus
from ax.analysis.healthcheck.no_effects_analysis import TestOfNoEffectAnalysis
from ax.core.data import Data
from ax.exceptions.core import AxError, UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
)


class TestTestOfNoEffectAnalysis(TestCase):
    def setUp(self) -> None:
        self.experiment = get_branin_experiment(with_trial=True)
        self.moo_experiment = get_branin_experiment_with_multi_objective(
            with_trial=True
        )
        self.tone = TestOfNoEffectAnalysis()

    def test_effects_detected(self) -> None:
        # GIVEN an experiment with effects detected
        self.experiment.attach_data(
            data=Data(
                df=pd.DataFrame(
                    {
                        "arm_name": ["0_0", "0_1", "0_2"],
                        "metric_name": ["branin"] * 3,
                        "mean": [0.0, 1.0, 2.0],
                        "sem": [0.1] * 3,
                        "trial_index": [0] * 3,
                        "n": [1000] * 3,
                    }
                )
            )
        )
        # WHEN we compute the healthcheck
        card = self.tone.compute(experiment=self.experiment)
        # THEN it is a PASS
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)
        self.assertEqual(card.name, "TestOfNoEffectAnalysis")
        self.assertEqual(card.title, "Ax Test of No Effect Success")
        self.assertEqual(
            "Effects are observed for all objective metrics.", card.subtitle
        )

    def test_no_effects_detected(self) -> None:
        # GIVEN an experiment with no effects detected
        self.experiment.attach_data(
            data=Data(
                df=pd.DataFrame(
                    {
                        "arm_name": ["0_0", "0_1", "0_2"],
                        "metric_name": ["branin"] * 3,
                        "mean": [1.0] * 3,  # same mean, so no effect detected
                        "sem": [0.1] * 3,
                        "trial_index": [0] * 3,
                        "n": [1000] * 3,
                    }
                )
            )
        )
        # WHEN we compute the healthcheck
        card = self.tone.compute(experiment=self.experiment)
        # THEN it is a WARNING
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)
        self.assertEqual(card.name, "TestOfNoEffectAnalysis")
        self.assertEqual(card.title, "Ax Test of No Effect Warning")
        self.assertIn("no effects have been detected", card.subtitle or "")

    def test_raises_error_with_no_experiment(self) -> None:
        # GIVEN no experiment is provided
        # WHEN we compute the healthcheck
        # THEN it raises a UserInputError
        with self.assertRaises(UserInputError):
            self.tone.compute(experiment=None)

    def test_raises_error_without_optimization_confi(self) -> None:
        # GIVEN we have an experiment with no opt conifg
        self.experiment._optimization_config = None
        # WHEN we compute the healthcheck
        # THEN it raises a UserInputError
        with self.assertRaises(UserInputError):
            self.tone.compute(experiment=self.experiment)

    def test_raises_error_without_data(self) -> None:
        # GIVEN we have an experiment with no data
        # WHEN we compute the healthcheck
        # THEN it raises an AxError
        with self.assertRaises(AxError):
            self.tone.compute(experiment=self.experiment)

    def test_multi_objective_partial_no_effects(self) -> None:
        # GIVEN we have a multi-objective experiment with one metric with no effects
        self.moo_experiment.attach_data(
            data=Data(
                df=pd.DataFrame(
                    {
                        "arm_name": ["status_quo", "0_0", "status_quo", "0_0"],
                        "metric_name": ["branin_a", "branin_a", "branin_b", "branin_b"],
                        "mean": [1.0, 1.0, 2.0, 4.0],
                        "sem": [0.1] * 4,
                        "trial_index": [0] * 4,
                        "n": [1000] * 4,
                    }
                )
            )
        )
        # WHEN we compute the healthcheck
        card = self.tone.compute(experiment=self.moo_experiment)
        # THEN it warns about one of the metrics
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)
        self.assertIn("branin_a", card.subtitle or "")
        self.assertNotIn("branin_b", card.subtitle or "")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationFeatures
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization
from ax.exceptions.core import UnsupportedError
from ax.generators.types import TConfig
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class FillMissingParameters(Transform):
    """If a parameter is missing from an arm, fill it with the value from
    the dict given in the config.

    Config supports two options.
        fill_values: a dict of {parameter_name: value} to fill in for missing
            parameters. Required.
        fill_None: a boolean indicating whether to fill in None values. Default
            is True. If False, parameters specified as None will remain None,
            and only parameters absent altogether will be filled.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        config = config or {}
        self.fill_values: TParameterization | None = config.get(  # pyre-ignore[8]
            "fill_values", None
        )
        self.fill_None: bool = assert_is_instance(config.get("fill_None", True), bool)

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        if self.fill_values is None:
            return observation_features
        for obsf in observation_features:
            fill_params = {
                k: v
                for k, v in none_throws(self.fill_values).items()
                if k not in obsf.parameters
                or (obsf.parameters[k] is None and self.fill_None)
            }
            obsf.parameters.update(fill_params)
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        if self.fill_values is None:
            return experiment_data
        if self.fill_None is False:
            # This shouldn't be relevant in regular usage. We add both
            # FillMissingParameters and Cast as default transfroms in
            # Adapter. Cast will drop parameterizations with missing / None
            # values, so not filling None will just lead to it being dropped.
            # The exception is added here for completeness.
            raise UnsupportedError(
                "Transforming `ExperimentData` is not supported for "
                "FillMissingParameters with fill_None=False. "
                "We cannot distinguish between parameters that are missing "
                "and those that are None in `ExperimentData`. "
            )
        arm_data = experiment_data.arm_data.fillna(value=self.fill_values)
        # If any of the fill columns are missing in arm_data, add it.
        missing_columns = set(none_throws(self.fill_values)) - set(arm_data.columns)
        for col in missing_columns:
            arm_data[col] = none_throws(self.fill_values)[col]
        return ExperimentData(
            arm_data=arm_data,
            observation_data=experiment_data.observation_data,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import json
import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any

from ax.adapter.base import Adapter
from ax.adapter.cross_validation import (
    compute_diagnostics,
    cross_validate,
    CVDiagnostics,
    CVResult,
    get_fit_and_std_quality_and_generalization_dict,
)
from ax.adapter.registry import GeneratorRegistryBase
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.exceptions.core import UserInputError
from ax.utils.common.base import SortableBase
from ax.utils.common.kwargs import consolidate_kwargs, get_function_argument_names
from ax.utils.common.serialization import SerializationMixin
from pyre_extensions import none_throws


class GeneratorSpecJSONEncoder(json.JSONEncoder):
    """Generic encoder to avoid JSON errors in GeneratorSpec.__repr__"""

    def default(self, o: Any) -> str:
        return repr(o)


@dataclass
class GeneratorSpec(SortableBase, SerializationMixin):
    generator_enum: GeneratorRegistryBase
    # Kwargs to pass into the `Adapter` + `Generator` constructors in
    # `GeneratorRegistryBase.__call__`.
    model_kwargs: dict[str, Any] = field(default_factory=dict)
    # Kwargs to pass to `Adapter.gen`.
    model_gen_kwargs: dict[str, Any] = field(default_factory=dict)
    # Kwargs to pass to `cross_validate`.
    model_cv_kwargs: dict[str, Any] = field(default_factory=dict)
    # An optional override for the model key. Each `GeneratorSpec` in a
    # `GenerationNode` must have a unique key to ensure identifiability.
    model_key_override: str | None = None

    # Fitted model, constructed using specified `model_kwargs` and `Data`
    # on `GeneratorSpec.fit`
    _fitted_adapter: Adapter | None = None

    # Stored cross validation results set in cross validate.
    _cv_results: list[CVResult] | None = None

    # Stored cross validation diagnostics set in cross validate.
    _diagnostics: CVDiagnostics | None = None

    # Stored to check if the CV result & diagnostic cache is safe to reuse.
    _last_cv_kwargs: dict[str, Any] | None = None

    # Stored to check if the model can be safely updated in fit.
    _last_fit_arg_ids: dict[str, int] | None = None

    def __post_init__(self) -> None:
        self.model_kwargs = self.model_kwargs or {}
        self.model_gen_kwargs = self.model_gen_kwargs or {}
        self.model_cv_kwargs = self.model_cv_kwargs or {}

    @property
    def fitted_adapter(self) -> Adapter:
        """Returns the fitted adapter, asserting fit() was called"""
        self._assert_fitted()
        return none_throws(self._fitted_adapter)

    @property
    def fixed_features(self) -> ObservationFeatures | None:
        """
        Fixed generation features to pass into the Model's `.gen` function.
        """
        return self.model_gen_kwargs.get("fixed_features", None)

    @fixed_features.setter
    def fixed_features(self, value: ObservationFeatures | None) -> None:
        """
        Fixed generation features to pass into the Model's `.gen` function.
        """
        self.model_gen_kwargs["fixed_features"] = value

    @property
    def model_key(self) -> str:
        """Key string to identify the model used by this ``GeneratorSpec``."""
        if self.model_key_override is not None:
            return self.model_key_override
        else:
            return self.generator_enum.value

    def fit(
        self,
        experiment: Experiment,
        data: Data | None = None,
        **model_kwargs: Any,
    ) -> None:
        """Fits the specified model on the given experiment + data using the
        model kwargs set on the model spec, alongside any passed down as
        kwargs to this function (local kwargs take precedent)
        """
        # unset any cross validation cache
        self._cv_results, self._diagnostics = None, None
        # NOTE: It's important to copy `self.model_kwargs` here to avoid actually
        # adding contents of `model_kwargs` passed to this method, to
        # `self.model_kwargs`.
        combined_model_kwargs = {**self.model_kwargs, **model_kwargs}
        if self._fitted_adapter is not None and self._safe_to_update(
            experiment=experiment, combined_model_kwargs=combined_model_kwargs
        ):
            # Update the data on the adapter and call `_fit`.
            # This will skip model fitting if the data has not changed.
            experiment_data, search_space = (
                self.fitted_adapter._process_and_transform_data(
                    experiment=experiment, data=data
                )
            )
            self.fitted_adapter._fit_if_implemented(
                search_space=search_space,
                experiment_data=experiment_data,
                time_so_far=0.0,
            )

        else:
            # Fit from scratch.
            self._fitted_adapter = self.generator_enum(
                experiment=experiment,
                data=data,
                **combined_model_kwargs,
            )
            self._last_fit_arg_ids = self._get_fit_arg_ids(
                experiment=experiment, combined_model_kwargs=combined_model_kwargs
            )

    def cross_validate(
        self,
        model_cv_kwargs: dict[str, Any] | None = None,
    ) -> tuple[list[CVResult] | None, CVDiagnostics | None]:
        """
        Call cross_validate, compute_diagnostics and cache the results.
        If the model cannot be cross validated, warn and return None.

        NOTE: If there are cached results, and the cache was computed using
        the same kwargs, this will return the cached results.

        Args:
            model_cv_kwargs: Optional kwargs to pass into `cross_validate` call.
                These are combined with `self.model_cv_kwargs`, with the
                `model_cv_kwargs` taking precedence over `self.model_cv_kwargs`.

        Returns:
            A tuple of CV results (observed vs predicted values) and the
            corresponding diagnostics.
        """
        cv_kwargs = {**self.model_cv_kwargs, **(model_cv_kwargs or {})}
        if (
            self._cv_results is not None
            and self._diagnostics is not None
            and cv_kwargs == self._last_cv_kwargs
        ):
            return self._cv_results, self._diagnostics

        self._assert_fitted()
        try:
            self._cv_results = cross_validate(model=self.fitted_adapter, **cv_kwargs)
        except NotImplementedError:
            warnings.warn(
                f"{self.generator_enum.value} cannot be cross validated", stacklevel=2
            )
            return None, None

        self._diagnostics = compute_diagnostics(self._cv_results)
        self._last_cv_kwargs = cv_kwargs
        return self._cv_results, self._diagnostics

    @property
    def cv_results(self) -> list[CVResult] | None:
        """
        Cached CV results from `self.cross_validate()`
        if it has been successfully called
        """
        return self._cv_results

    @property
    def diagnostics(self) -> CVDiagnostics | None:
        """
        Cached CV diagnostics from `self.cross_validate()`
        if it has been successfully called
        """
        return self._diagnostics

    def gen(self, **model_gen_kwargs: Any) -> GeneratorRun:
        """Generates candidates from the fitted model, using the model gen
        kwargs set on the model spec, alongside any passed as kwargs
        to this function (local kwargs take precedent)

        NOTE: Model must have been fit prior to calling gen()

        Args:
            n: Integer representing how many arms should be in the generator run
                produced by this method. NOTE: Some underlying models may ignore
                the ``n`` and produce a model-determined number of arms. In that
                case this method will also output a generator run with number of
                arms that can differ from ``n``.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
        """
        fitted_adapter = self.fitted_adapter
        model_gen_kwargs = consolidate_kwargs(
            kwargs_iterable=[self.model_gen_kwargs, model_gen_kwargs],
            keywords=get_function_argument_names(fitted_adapter.gen),
        )
        # copy to ensure there is no in-place modification
        model_gen_kwargs = deepcopy(model_gen_kwargs)
        generator_run = fitted_adapter.gen(**model_gen_kwargs)
        fit_and_std_quality_and_generalization_dict = (
            get_fit_and_std_quality_and_generalization_dict(
                fitted_adapter=self.fitted_adapter,
            )
        )
        generator_run._gen_metadata = (
            {} if generator_run.gen_metadata is None else generator_run.gen_metadata
        )
        generator_run._gen_metadata.update(
            **fit_and_std_quality_and_generalization_dict
        )
        return generator_run

    def copy(self) -> GeneratorSpec:
        """`GeneratorSpec` is both a spec and an object that performs actions.
        Copying is useful to avoid changes to a singleton model spec.
        """
        return self.__class__(
            generator_enum=self.generator_enum,
            model_kwargs=deepcopy(self.model_kwargs),
            model_gen_kwargs=deepcopy(self.model_gen_kwargs),
            model_cv_kwargs=deepcopy(self.model_cv_kwargs),
            model_key_override=self.model_key_override,
        )

    def _safe_to_update(
        self,
        experiment: Experiment,
        combined_model_kwargs: dict[str, Any],
    ) -> bool:
        """Checks if the object id of any of the non-data fit arguments has changed.

        This is a cheap way of checking that we're attempting to re-fit the same
        model for the same experiment, which is a very reasonable expectation
        since this all happens on the same `GeneratorSpec` instance.
        """
        if self.model_key == "TRBO":
            # Temporary hack to unblock TRBO.
            # TODO[T167756515] Remove when TRBO revamp diff lands.
            return True
        return self._last_fit_arg_ids == self._get_fit_arg_ids(
            experiment=experiment, combined_model_kwargs=combined_model_kwargs
        )

    def _get_fit_arg_ids(
        self,
        experiment: Experiment,
        combined_model_kwargs: dict[str, Any],
    ) -> dict[str, int]:
        """Construct a dictionary mapping arg name to object id."""
        return {
            "experiment": id(experiment),
            **{k: id(v) for k, v in combined_model_kwargs.items()},
        }

    def _assert_fitted(self) -> None:
        """Helper that verifies a model was fitted, raising an error if not"""
        if self._fitted_adapter is None:
            raise UserInputError("No fitted model found. Call fit() to generate one")

    def _brief_repr(self) -> str:
        """Returns a brief string representation of this model spec.
        Includes just name and override, but not the various kwargs"""
        return (
            "GeneratorSpec("
            f"\tgenerator_enum={self.generator_enum.value}, "
            f"\tmodel_key_override={self.model_key_override}"
            ")"
        )

    def __repr__(self) -> str:
        model_kwargs = json.dumps(
            self.model_kwargs, sort_keys=True, cls=GeneratorSpecJSONEncoder
        )
        model_gen_kwargs = json.dumps(
            self.model_gen_kwargs, sort_keys=True, cls=GeneratorSpecJSONEncoder
        )
        model_cv_kwargs = json.dumps(
            self.model_cv_kwargs, sort_keys=True, cls=GeneratorSpecJSONEncoder
        )
        return (
            "GeneratorSpec("
            f"\tgenerator_enum={self.generator_enum.value}, "
            f"\tmodel_kwargs={model_kwargs}, "
            f"\tmodel_gen_kwargs={model_gen_kwargs}, "
            f"\tmodel_cv_kwargs={model_cv_kwargs}, "
            f"\tmodel_key_override={self.model_key_override}"
            ")"
        )

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: GeneratorSpec) -> bool:
        return repr(self) == repr(other)

    @property
    def _unique_id(self) -> str:
        """Returns the unique ID of this model spec"""
        # TODO @mgarrard verify that this is unique enough
        return str(hash(self))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from collections.abc import Sequence
from dataclasses import dataclass
from unittest import mock

import numpy as np

import pandas as pd
from ax.core.map_data import MapData
from ax.core.metric import MetricFetchE
from ax.metrics.tensorboard import logger, TensorboardMetric
from ax.utils.common.result import Ok
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_trial
from pyre_extensions import assert_is_instance
from tensorboard.backend.event_processing import event_multiplexer


@dataclass
class _TensorProto:
    double_val: list[float]


@dataclass
class _TensorEvent:
    step: int
    tensor_proto: _TensorProto


def _get_fake_multiplexer(
    fake_data: Sequence[float],
) -> event_multiplexer.EventMultiplexer:
    mul = event_multiplexer.EventMultiplexer()

    # pyre-ignore[8] Return fake tags when content is requested
    mul.PluginRunToTagToContent = lambda plugin: {".": {"loss": ""}}

    # pyre-ignore[8] Return fake data when tensors requested
    mul.Tensors = lambda run, tag: [
        _TensorEvent(step=i, tensor_proto=_TensorProto(double_val=[dat]))
        for i, dat in enumerate(fake_data)
    ]

    return mul


def tb_smooth(scalars: list[float], weight: float) -> list[float]:
    """
    Debiased EMA implementation according to Tensorboard Github:
    https://fburl.com/6v6q8scg

    This function is used to test that the smoothing functaionality
    of pandas.series.ewm is equivalent to the smoothing functionality
    of Tensorboard.
    """
    last = 0
    smoothed = []
    num_acc = 0
    for next_val in scalars:
        last = last * weight + (1 - weight) * next_val
        num_acc += 1
        # de-bias
        debias_weight = 1
        if weight != 1:
            debias_weight = 1 - math.pow(weight, num_acc)
        smoothed_val = last / debias_weight
        smoothed.append(smoothed_val)

    return smoothed


class TensorboardMetricTest(TestCase):
    def test_fetch_trial_data(self) -> None:
        fake_data = [8.0, 9.0, 2.0, 1.0]
        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)
        metric = TensorboardMetric(
            name="loss", tag="loss", lower_is_better=True, smoothing=0
        )
        trial = get_trial()

        # This is mocked to avoid an xdb call
        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            result = metric.fetch_trial_data(trial=trial)

        df = assert_is_instance(result.unwrap(), MapData).map_df

        expected_df = pd.DataFrame(
            [
                {
                    "trial_index": 0,
                    "arm_name": "0_0",
                    "metric_name": "loss",
                    "mean": fake_data[i],
                    "sem": float("nan"),
                    "step": float(i),
                }
                for i in range(len(fake_data))
            ]
        )

        self.assertTrue(df.equals(expected_df))

    def test_fetch_trial_data_with_bad_data(self) -> None:
        nan_data = [1, 2, np.nan, 4]
        nan_multiplexer = _get_fake_multiplexer(fake_data=nan_data)
        metric = TensorboardMetric(name="loss", tag="loss")

        trial = get_trial()
        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=nan_multiplexer,
        ), mock.patch.object(logger, "warning") as mock_warning:
            result = metric.fetch_trial_data(trial=trial)
        mock_warning.assert_called_once_with(
            "1 / 4 data points are NaNs or Infs. Filtering out non-finite values."
        )

        assert_is_instance(result, Ok)
        map_data = assert_is_instance(result.unwrap(), MapData)
        nan_array = np.array(nan_data)
        nan_array = nan_array[np.isfinite(nan_array)]
        self.assertTrue(
            np.array_equal(
                map_data.map_df["mean"].to_numpy(), nan_array, equal_nan=False
            )
        )

        # testing some inf data
        inf_data = [1, 2, np.inf, 4]
        inf_multiplexer = _get_fake_multiplexer(fake_data=inf_data)

        metric = TensorboardMetric(name="loss", tag="loss")

        trial = get_trial()
        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=inf_multiplexer,
        ):
            with mock.patch.object(logger, "warning") as mock_warning:
                result = metric.fetch_trial_data(trial=trial)
            mock_warning.assert_called_once_with(
                "1 / 4 data points are NaNs or Infs. Filtering out non-finite values."
            )

        assert_is_instance(result, Ok)
        map_data = assert_is_instance(result.unwrap(), MapData)
        inf_array = np.array(inf_data)
        inf_array = inf_array[np.isfinite(inf_array)]
        self.assertTrue(np.array_equal(map_data.map_df["mean"].to_numpy(), inf_array))

        # testing all non-finite data
        for non_finite_val in [np.nan, np.inf]:
            nf_data = [non_finite_val for _ in range(4)]
            nf_multiplexer = _get_fake_multiplexer(fake_data=nf_data)
            metric = TensorboardMetric(name="loss", tag="loss")

            trial = get_trial()
            with mock.patch.object(
                TensorboardMetric,
                "_get_event_multiplexer_for_trial",
                return_value=nf_multiplexer,
            ):
                result = metric.fetch_trial_data(trial=trial)

            err = assert_is_instance(result.unwrap_err(), MetricFetchE)
            self.assertEqual(err.message, "Failed to fetch data for loss")
            self.assertEqual(str(err.exception), "All values are NaNs or Infs.")

    def test_smoothing(self) -> None:
        fake_data = [8.0, 4.0, 2.0, 1.0]
        smoothing = 0.5
        smooth_data = tb_smooth(fake_data, smoothing)

        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)
        metric = TensorboardMetric(
            name="loss", tag="loss", lower_is_better=True, smoothing=smoothing
        )
        trial = get_trial()

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            result = metric.fetch_trial_data(trial=trial)

        df = assert_is_instance(result.unwrap(), MapData).map_df

        expected_df = pd.DataFrame(
            [
                {
                    "trial_index": 0,
                    "arm_name": "0_0",
                    "metric_name": "loss",
                    "mean": smooth_data[i],
                    "sem": float("nan"),
                    "step": float(i),
                }
                for i in range(len(fake_data))
            ]
        )

        self.assertTrue(df.equals(expected_df))

    def test_cumulative_best(self) -> None:
        fake_data = [4.0, 8.0, 2.0, 1.0]
        cummin_data = pd.Series(fake_data).cummin().tolist()

        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)
        metric = TensorboardMetric(
            name="loss",
            tag="loss",
            lower_is_better=True,
            cumulative_best=True,
            smoothing=0,
        )
        trial = get_trial()

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            result = metric.fetch_trial_data(trial=trial)

        df = assert_is_instance(result.unwrap(), MapData).map_df

        expected_df = pd.DataFrame(
            [
                {
                    "trial_index": 0,
                    "arm_name": "0_0",
                    "metric_name": "loss",
                    "mean": cummin_data[i],
                    "sem": float("nan"),
                    "step": float(i),
                }
                for i in range(len(fake_data))
            ]
        )

        self.assertTrue(df.equals(expected_df))

    def test_percentile(self) -> None:
        fake_data = [8.0, 4.0, 2.0, 1.0]
        percentile = 0.5
        percentile_data = pd.Series(fake_data).expanding().quantile(percentile)
        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)
        trial = get_trial()

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            metric = TensorboardMetric(
                name="loss", tag="loss", lower_is_better=True, percentile=percentile
            )
            result = metric.fetch_trial_data(trial=trial)
        df = assert_is_instance(result.unwrap(), MapData).map_df
        expected_df = pd.DataFrame(
            [
                {
                    "trial_index": 0,
                    "arm_name": "0_0",
                    "metric_name": "loss",
                    "mean": percentile_data[i],
                    "sem": float("nan"),
                    "step": float(i),
                }
                for i in range(len(fake_data))
            ]
        )

        self.assertTrue(df.equals(expected_df))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
import json

import torch
from ax.adapter.registry import Generators

from ax.analysis.plotly.objective_p_feasible_frontier import (
    ObjectivePFeasibleFrontierPlot,
)
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ScalarizedOutcomeConstraint
from ax.core.types import ComparisonOp
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment_with_multi_objective,
    get_branin_metric,
)
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.utils.testing import skip_if_import_error
from pyre_extensions import none_throws


class TestObjectivePFeasibleFrontierPlot(TestCase):
    def setUp(self) -> None:
        super().setUp()

        torch.manual_seed(0)
        self.experiment = get_branin_experiment_with_multi_objective(
            with_completed_batch=True, with_absolute_constraint=True, num_objectives=3
        )
        self.experiment.optimization_config = OptimizationConfig(
            objective=Objective(metric=self.experiment.metrics["branin_a"]),
            outcome_constraints=self.experiment.optimization_config.outcome_constraints,
        )
        oc = none_throws(self.experiment.optimization_config).outcome_constraints[0]
        oc.bound = 10.0
        oc.op = ComparisonOp.LEQ

    @skip_if_import_error
    @mock_botorch_optimize
    def test_compute(self) -> None:
        # import pymoo here to raise an import error if pymoo is not installed.
        # The import error is already handled in
        # ax/generators/torch/botorch_modular/acquisition.py so we need to
        # reimport it here
        import pymoo  # noqa: F401

        adapter = Generators.BOTORCH_MODULAR(experiment=self.experiment)
        card = ObjectivePFeasibleFrontierPlot().compute(
            experiment=self.experiment, adapter=adapter
        )
        self.assertEqual(
            json.loads(card.blob)["layout"]["xaxis"]["title"]["text"], "branin_a"
        )
        self.assertEqual(
            json.loads(card.blob)["layout"]["yaxis"]["title"]["text"],
            "% Chance of Satisfying the Constraints",
        )
        self.assertFalse(card.df["branin_a_sem"].isna().any())
        self.assertTrue(card.df["p_feasible_sem"].isna().all())

    def test_no_exceptions(self) -> None:
        with self.assertRaisesRegex(
            UserInputError, "ObjectivePFeasibleFrontierPlot requires an Experiment."
        ):
            ObjectivePFeasibleFrontierPlot().compute(experiment=None)
        opt_config = self.experiment.optimization_config
        self.experiment._optimization_config = None
        with self.assertRaisesRegex(
            UserInputError, "Optimization_config must be set to compute frontier."
        ):
            ObjectivePFeasibleFrontierPlot().compute(experiment=self.experiment)
        self.experiment.optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(
                objectives=[
                    Objective(metric=m) for m in self.experiment.metrics.values()
                ]
            )
        )
        with self.assertRaisesRegex(
            UnsupportedError, "Multi-objective optimization is not supported."
        ):
            ObjectivePFeasibleFrontierPlot().compute(experiment=self.experiment)
        self.experiment.optimization_config = opt_config
        opt_config.outcome_constraints = []
        with self.assertRaisesRegex(
            UserInputError,
            r"Plotting the objective-p\(feasible\) frontier requires at least one "
            "outcome constraint.",
        ):
            ObjectivePFeasibleFrontierPlot().compute(experiment=self.experiment)
        self.experiment.add_tracking_metric(get_branin_metric("branin2"))
        # Get only tracking metrics, excluding the objective metric to avoid
        # "Cannot constrain on objective metric" error
        constraint_metrics = [
            self.experiment.metrics["branin_b"],
            self.experiment.metrics["branin_c"],
        ]
        opt_config.outcome_constraints = [
            ScalarizedOutcomeConstraint(
                metrics=constraint_metrics,
                weights=[1.0, 1.0],
                relative=False,
                bound=10.0,
                op=ComparisonOp.LEQ,
            )
        ]
        with self.assertRaisesRegex(
            UnsupportedError,
            "Scalarized outcome constraints are not supported yet.",
        ):
            ObjectivePFeasibleFrontierPlot().compute(experiment=self.experiment)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.adapter.factory import get_thompson
from ax.analysis.plotly.marginal_effects import (
    compute_marginal_effects_adhoc,
    MarginalEffectsPlot,
)
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_discrete_search_space, get_sobol


class TestMarginalEffectsPlot(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.experiment = Experiment(
            search_space=get_discrete_search_space(),
            name="test_experiment",
            is_test=True,
            optimization_config=OptimizationConfig(
                objective=Objective(Metric("metric_1", lower_is_better=False))
            ),
        )
        num_arms = 3
        num_trials = 3
        sobol = get_sobol(search_space=self.experiment.search_space)
        gr = sobol.gen(n=num_arms)

        for i in range(num_trials):
            self.experiment.new_batch_trial(generator_run=gr)
            self.experiment.trials[i].mark_running(no_runner_required=True)
            self.experiment.attach_data(
                Data(
                    pd.DataFrame(
                        {
                            "trial_index": [i] * num_arms,
                            "arm_name": [f"0_{j}" for j in range(num_arms)],
                            "metric_name": ["metric_1"] * num_arms,
                            "mean": list(range(num_arms)),
                            "sem": [0.5] * num_arms,
                        }
                    )
                )
            )

        self.analysis_all_variables = MarginalEffectsPlot(
            metric_name="metric_1", parameters=["x", "y", "z"]
        )
        self.analysis_default_params = MarginalEffectsPlot(metric_name="metric_1")

    def test_compute(self) -> None:
        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            self.analysis_default_params.compute()

        adapter = get_thompson(
            experiment=self.experiment,
            data=self.experiment.lookup_data(trial_indices=[0]),
        )

        with self.assertRaisesRegex(
            UserInputError, "MarginalEffectsPlot is only for `ChoiceParameter`s"
        ):
            self.analysis_all_variables.compute(
                experiment=self.experiment, adapter=adapter
            )

        cards = self.analysis_default_params.compute(
            experiment=self.experiment, adapter=adapter
        ).flatten()

        self.assertEqual(cards[0].name, "MarginalEffectsPlot")
        self.assertEqual(
            cards[0].title, "Marginal Effects for z"
        )  # z is the only choice parameter
        self.assertEqual(
            cards[0].subtitle,
            (
                "This plot visualizes the predicted relative changes in "
                "metric_1 for each factor level of z, offering insights "
                "into their impact on the metric. By comparing the "
                "effects of different levels, this plot helps identify which "
                "factor levels have the most significant influence, providing "
                "a detailed understanding of the experimental results and guiding "
                "future decisions on factor selection."
            ),
        )
        self.assertIsNotNone(cards[0].df)
        self.assertIsNotNone(cards[0].blob)
        self.assertEqual(len(cards), 1)
        self.assertEqual(
            set(cards[0].df.columns),
            {"Name", "Level", "Beta", "SE"},
        )

    def test_compute_adhoc(self) -> None:
        adapter = get_thompson(
            experiment=self.experiment,
            data=self.experiment.lookup_data(trial_indices=[0]),
        )

        card = compute_marginal_effects_adhoc(
            metric_name="metric_1",
            experiment=self.experiment,
            adapter=adapter,
        ).flatten()

        self.assertEqual(card[0].name, "MarginalEffectsPlot")
        self.assertEqual(
            card[0].title, "Marginal Effects for z"
        )  # Assuming 'z' is the only choice parameter
        self.assertEqual(
            card[0].subtitle,
            (
                "This plot visualizes the predicted relative changes in "
                "metric_1 for each factor level of z, offering insights "
                "into their impact on the metric. By comparing the "
                "effects of different levels, this plot helps identify which "
                "factor levels have the most significant influence, providing "
                "a detailed understanding of the experimental results and guiding "
                "future decisions on factor selection."
            ),
        )
        self.assertIsNotNone(card[0].df)
        self.assertIsNotNone(card[0].blob)
        self.assertEqual(len(card), 1)
        self.assertEqual(
            set(card[0].df.columns),
            {"Name", "Level", "Beta", "SE"},
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

from ax.adapter.cross_validation import cross_validate
from ax.adapter.registry import Generators
from ax.analysis.plotly.cross_validation import (
    compute_cross_validation_adhoc,
    CrossValidationPlot,
)
from ax.core.trial import Trial
from ax.exceptions.core import UserInputError
from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_default_generation_strategy_at_MBM_node
from pyre_extensions import assert_is_instance, none_throws


class TestCrossValidationPlot(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        self.client = AxClient()
        self.client.create_experiment(
            is_test=True,
            name="foo",
            parameters=[
                {
                    "name": "x",
                    "type": "range",
                    "bounds": [-1.0, 1.0],
                }
            ],
            objectives={"bar": ObjectiveProperties(minimize=True)},
            tracking_metric_names=["foo"],
        )

        for _ in range(10):
            parameterization, trial_index = self.client.get_next_trial()
            self.client.complete_trial(
                trial_index=trial_index,
                raw_data={
                    "bar": parameterization["x"] ** 2,
                    "foo": parameterization["x"] ** 3,
                },
            )

    def test_compute(self) -> None:
        analysis = CrossValidationPlot(metric_names=["bar"])

        # Test that it fails if no GenerationStrategy is provided
        with self.assertRaisesRegex(
            UserInputError, "Must provide either a GenerationStrategy or an Adapter"
        ):
            analysis.compute()

        (card,) = analysis.compute(
            generation_strategy=self.client.generation_strategy
        ).flatten()
        self.assertEqual(
            card.name,
            "CrossValidationPlot",
        )
        self.assertEqual(card.title, "Cross Validation for bar")
        self.assertEqual(
            card.subtitle,
            (
                "The cross-validation plot displays the model fit for each "
                "metric in the experiment. It employs a leave-one-out "
                "approach, where the model is trained on all data except one "
                "sample, which is used for validation. The plot shows the "
                "predicted outcome for the validation set on the y-axis against "
                "its actual value on the x-axis. Points that align closely with "
                "the dotted diagonal line indicate a strong model fit, signifying "
                "accurate predictions. Additionally, the plot includes 95% "
                "confidence intervals that provide insight into the noise in "
                "observations and the uncertainty in model predictions. A "
                "horizontal, flat line of predictions indicates that the model "
                "has not picked up on sufficient signal in the data, and instead "
                "is just predicting the mean."
            ),
        )
        self.assertEqual(
            {*card.df.columns},
            {"arm_name", "observed", "observed_95_ci", "predicted", "predicted_95_ci"},
        )
        self.assertIsNotNone(card.blob)

        # Assert that all arms are in the cross validation df
        # because trial index is not specified
        for t in self.client.experiment.trials.values():
            # Skip the last trial because the model was used to generate it
            # and therefore hasn't observed it
            if t.index == max(self.client.experiment.trials.keys()):
                continue
            arm_name = none_throws(assert_is_instance(t, Trial).arm).name
            self.assertIn(
                arm_name,
                card.df["arm_name"].unique(),
            )

    def test_it_can_specify_trial_index_correctly(self) -> None:
        analysis = CrossValidationPlot(metric_names=["bar"], trial_index=9)
        (card,) = analysis.compute(
            generation_strategy=self.client.generation_strategy
        ).flatten()
        for t in self.client.experiment.trials.values():
            # Skip the last trial because the model was used to generate it
            # and therefore hasn't observed it
            if t.index == max(self.client.experiment.trials.keys()):
                continue
            arm_name = none_throws(assert_is_instance(t, Trial).arm).name
            self.assertIn(
                arm_name,
                card.df["arm_name"].unique(),
            )

    @mock.patch(
        "ax.analysis.plotly.cross_validation.cross_validate", wraps=cross_validate
    )
    def test_cross_validate_is_called_once_with_multiple_metrics(
        self, mock_cross_validate: mock.Mock
    ) -> None:
        analysis = CrossValidationPlot()
        analysis.compute(generation_strategy=self.client.generation_strategy)
        mock_cross_validate.assert_called_once()

    @mock_botorch_optimize
    def test_compute_adhoc(self) -> None:
        metric_mapping = {"bar": "spunky", "foo": "foo2"}
        data = self.client.experiment.lookup_data()
        adapter = Generators.BOTORCH_MODULAR(
            experiment=self.client.experiment, data=data
        )
        cards = compute_cross_validation_adhoc(
            adapter=adapter, labels=metric_mapping
        ).flatten()
        self.assertEqual(len(cards), 2)
        titles = {"Cross Validation for spunky", "Cross Validation for foo2"}
        for card in cards:
            self.assertEqual(card.name, "CrossValidationPlot")
            # validate that the metric name replacement occurred
            self.assertIn(card.title, titles)
            titles.remove(card.title)

    @TestCase.ax_long_test(
        reason=(
            "cross_validate still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_online(self) -> None:
        # Test CrossValidationPlot can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        for experiment in get_online_experiments():
            for untransform in [True, False]:
                for refined_metric_name in [None, "foo"]:
                    generation_strategy = get_default_generation_strategy_at_MBM_node(
                        experiment=experiment
                    )

                    # Pick an arbitrary metric from the experiment's optimization config
                    metric_name = none_throws(
                        experiment.optimization_config
                    ).objective.metric_names[0]

                    analysis = CrossValidationPlot(
                        metric_names=[metric_name],
                        untransform=untransform,
                        labels={metric_name: refined_metric_name}
                        if refined_metric_name
                        else None,
                    )

                    _ = analysis.compute(
                        experiment=experiment, generation_strategy=generation_strategy
                    )

    @TestCase.ax_long_test(
        reason=(
            "cross_validate still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_offline(self) -> None:
        # Test CrossValidationPlot can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        for experiment in get_offline_experiments():
            for untransform in [True, False]:
                for refined_metric_name in [None, "foo"]:
                    generation_strategy = get_default_generation_strategy_at_MBM_node(
                        experiment=experiment
                    )

                    # Pick an arbitrary metric from the experiment's optimization config
                    metric_name = none_throws(
                        experiment.optimization_config
                    ).objective.metric_names[0]

                    analysis = CrossValidationPlot(
                        metric_names=[metric_name],
                        untransform=untransform,
                        labels={metric_name: refined_metric_name}
                        if refined_metric_name
                        else None,
                    )

                    _ = analysis.compute(
                        experiment=experiment, generation_strategy=generation_strategy
                    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.deprecated_transform_mixin import DeprecatedTransformMixin
from ax.adapter.transforms.utils import ClosestLookupDict, construct_new_search_space
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ChoiceParameter, Parameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import TParamValue
from ax.generators.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class ChoiceToNumericChoice(Transform):
    """Convert non-numeric ChoiceParameters to integer ChoiceParameters.

    If the parameter type is numeric (int, float), the parameter is not modified.
    If the parameter is non-numeric, this transform uses an integer encoding to
    `0, 1, ..., n_choices - 1`. The resulting choice parameter will be considered
    ordered iff the original parameter is.

    In the inverse transform, parameters will be mapped back onto the original domain.

    This transform does not transform task parameters
    (use TaskChoiceToIntTaskChoice for this).

    Note that this behavior is different from that of OrderedChoiceToIntegerRange, which
    transforms (ordered) ChoiceParameters to integer RangeParameters (rather than
    ChoiceParameters).

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        assert search_space is not None, "ChoiceToNumericChoice requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Identify parameters that should be transformed
        self.encoded_parameters: dict[str, dict[TParamValue, TParamValue]] = {}
        self.encoded_parameters_inverse: dict[str, ClosestLookupDict] = {}
        for p in search_space.parameters.values():
            if isinstance(p, ChoiceParameter) and not p.is_numeric and not p.is_task:
                transformed_values = list(range(len(p.values)))
                self.encoded_parameters[p.name] = dict(
                    zip(p.values, transformed_values)
                )
                self.encoded_parameters_inverse[p.name] = ClosestLookupDict(
                    zip(transformed_values, p.values)
                )

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.encoded_parameters:
                if p_name in obsf.parameters:
                    obsf.parameters[p_name] = self.encoded_parameters[p_name][
                        obsf.parameters[p_name]
                    ]
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        transformed_parameters: dict[str, Parameter] = {}
        for p_name, p in search_space.parameters.items():
            if p_name in self.encoded_parameters and isinstance(p, ChoiceParameter):
                encoding = self.encoded_parameters[p_name]
                dependents = None
                if p.is_hierarchical:
                    # The dependents of hierarchical parameters need to be updated to
                    # reflect the changes by encoding.
                    dependents = {
                        encoding[val]: deps for val, deps in p.dependents.items()
                    }

                transformed_parameters[p_name] = ChoiceParameter(
                    name=p_name,
                    parameter_type=ParameterType.INT,
                    # Explicitly extracting values rather than passing
                    # `list(encoding.values())` should make it possible to correctly
                    # transform a parameter that has different (a subset of) values
                    # than the parameter used to initialize the transform.
                    values=[encoding[val] for val in p.values],
                    is_ordered=p.is_ordered,
                    is_task=p.is_task,
                    is_fidelity=p.is_fidelity,
                    target_value=encoding[p.target_value]
                    if p.target_value is not None
                    else None,
                    sort_values=p.sort_values,
                    dependents=dependents,
                )
            else:
                transformed_parameters[p.name] = p
        return construct_new_search_space(
            search_space=search_space,
            parameters=list(transformed_parameters.values()),
            parameter_constraints=[
                pc.clone_with_transformed_parameters(
                    transformed_parameters=transformed_parameters
                )
                for pc in search_space.parameter_constraints
            ],
        )

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name, reverse_transform in self.encoded_parameters_inverse.items():
                if p_name in obsf.parameters:
                    # pyre: pval is declared to have type `int` but is used as
                    # pyre-fixme[9]: type `Union[bool, float, str]`.
                    pval: int = obsf.parameters[p_name]
                    if pval in reverse_transform:
                        obsf.parameters[p_name] = reverse_transform[pval]
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        return ExperimentData(
            arm_data=experiment_data.arm_data.replace(
                to_replace=self.encoded_parameters
            ),
            observation_data=experiment_data.observation_data,
        )


class ChoiceEncode(DeprecatedTransformMixin, ChoiceToNumericChoice):
    """Deprecated alias for ChoiceToNumericChoice."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class OrderedChoiceToIntegerRange(ChoiceToNumericChoice):
    """Convert ordered ChoiceParameters to integer RangeParameters.

    Parameters will be transformed to an integer RangeParameters, mapped from the
    original choice domain to a contiguous range `0, 1, ..., n_choices - 1`
    of integers. Does not transform task parameters.

    In the inverse transform, parameters will be mapped back onto the original domain.

    In order to encode all ChoiceParameters (not just ordered ChoiceParameters),
    use ChoiceToNumericChoice instead.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: SearchSpace,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Identify parameters that should be transformed
        self.encoded_parameters: dict[str, dict[TParamValue, int]] = {}
        for p in search_space.parameters.values():
            if isinstance(p, ChoiceParameter) and p.is_ordered and not p.is_task:
                self.encoded_parameters[p.name] = {
                    original_value: transformed_value
                    for transformed_value, original_value in enumerate(p.values)
                }
        self.encoded_parameters_inverse: dict[str, dict[int, TParamValue]] = {
            p_name: {
                transformed_value: original_value
                for original_value, transformed_value in transforms.items()
            }
            for p_name, transforms in self.encoded_parameters.items()
        }

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        transformed_parameters: dict[str, Parameter] = {}
        for p_name, p in search_space.parameters.items():
            if p_name in self.encoded_parameters and isinstance(p, ChoiceParameter):
                if p.is_fidelity:
                    raise ValueError(
                        f"Cannot choice-encode fidelity parameter {p_name}"
                    )
                # Make sure that the search space is compatible with the encoding.
                encoding = self.encoded_parameters[p_name]
                try:
                    t_values = [encoding[pv] for pv in p.values]
                except KeyError:
                    raise ValueError(
                        f"The parameter {p} contains values that are not present in "
                        "the search space used to initialize the transform. The "
                        f"supported encoding for the parameter {p_name} is {encoding}."
                    )
                min_val = min(t_values)
                len_val = len(t_values)
                # Ensure that the values span a contiguous range.
                if set(t_values) != set(range(min_val, min_val + len_val)):
                    raise ValueError(
                        f"The {self.__class__.__name__} transform requires the "
                        "parameter to be encoded with a contiguous range of integers. "
                        f"The parameter {p} maps to {t_values}, which does not span "
                        f"a contiguous range of integers. For parameter {p_name}, "
                        f"the transform uses {encoding=}."
                    )
                transformed_parameters[p_name] = RangeParameter(
                    name=p_name,
                    parameter_type=ParameterType.INT,
                    lower=min_val,
                    upper=min_val + len_val - 1,
                )
            else:
                transformed_parameters[p.name] = p
        return construct_new_search_space(
            search_space=search_space,
            parameters=list(transformed_parameters.values()),
            parameter_constraints=[
                pc.clone_with_transformed_parameters(
                    transformed_parameters=transformed_parameters
                )
                for pc in search_space.parameter_constraints
            ],
        )


class OrderedChoiceEncode(DeprecatedTransformMixin, OrderedChoiceToIntegerRange):
    """Deprecated alias for OrderedChoiceToIntegerRange."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from typing import Literal, Mapping, Sequence

import pandas as pd
from ax.adapter.base import Adapter
from ax.adapter.torch import TorchAdapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.plotly.color_constants import COLOR_FOR_DECREASES, COLOR_FOR_INCREASES
from ax.analysis.plotly.plotly_analysis import create_plotly_analysis_card
from ax.analysis.plotly.utils import (
    LEGEND_POSITION,
    MARGIN_REDUCUTION,
    MAX_HOVER_LABEL_LEN,
    truncate_label,
)
from ax.analysis.utils import extract_relevant_adapter
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.sensitivity.sobol_measures import ax_parameter_sens
from plotly import express as px, graph_objects as go
from pyre_extensions import override

# SensitivityAnalysisPlot uses a plotly bar chart which needs especially short labels
MAX_LABEL_LEN: int = 20

SENSITIVITY_CARDGROUP_TITLE = (
    "Sensitivity Analysis: Understand how each parameter affects metrics"
)

SENSITIVITY_CARDGROUP_SUBTITLE = (
    "These plots showcase the most influential parameters for each metric in the "
    "experiment, highlighting both the direction and magnitude of the metric's "
    "sensitivity to changes in these parameters. This information can be valuable for "
    "understanding metrics that may be oppositely affected by the same parameter, "
    "identifying the most critical parameters to further refine the search space, or "
    "validating underlying assumptions about the experiment's response surface. "
    "Sensitivity is measured using Sobol indices, which are calculated based on the "
    "model fitted to the data."
)


class SensitivityAnalysisPlot(Analysis):
    """
    Compute sensitivity for all metrics on a TorchAdapter.

    Sobol measures are always positive regardless of the direction in which the
    parameter influences f. If `signed` is set to True, then the Sobol measure for each
    parameter will be given as its sign the sign of the average gradient with respect to
    that parameter across the search space. Thus, important parameters that, when
    increased, decrease will have large and negative values; unimportant parameters
    will have values close to 0.
    """

    def __init__(
        self,
        metric_names: Sequence[str] | None = None,
        order: Literal["first", "second", "total"] = "total",
        top_k: int | None = 6,
        labels: Mapping[str, str] | None = None,
    ) -> None:
        """
        Args:
            metric_names: The names of the metrics and outcomes for which to compute
                sensitivities. This should preferably be metrics with a good model fit.
                Defaults to all metrics in the experiment.
            order: A string specifying the order of the Sobol indices to be computed.
                Supports "first" and "total" and defaults to "first".
            top_k: Optional limit on the number of parameters to show in the plot.
            labels: A mapping from metric names to labels to use in the plot. If a label
                is not provided for a metric, the metric name will be used.
        """
        self.metric_names = metric_names
        self.order = order
        self.top_k = top_k
        self.labels: dict[str, str] = {**labels} if labels is not None else {}

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        if not isinstance(relevant_adapter, TorchAdapter):
            raise UserInputError(
                "SensitivityAnalysisPlot requires a TorchAdapter, found "
                f"{type(adapter)}."
            )

        data = _prepare_data(
            adapter=relevant_adapter,
            metric_names=self.metric_names,
            order=self.order,
        )

        cards = []
        for metric_name in data["metric_name"].unique():
            # If a human readable metric name is provided, use it
            metric_label = self.labels.get(
                metric_name, truncate_label(label=metric_name, n=MAX_LABEL_LEN)
            )
            df, fig = _prepare_card_components(
                data=data,
                metric_name=metric_name,
                top_k=self.top_k,
                metric_label=metric_label,
            )

            card = create_plotly_analysis_card(
                name=self.__class__.__name__,
                title=f"Sensitivity Analysis for {metric_label}",
                subtitle=(
                    f"Understand how each parameter affects {metric_label} according "
                    f"to a {self.order}-order sensitivity analysis."
                ),
                df=df,
                fig=fig,
            )

            cards.append(card)

        return self._create_analysis_card_group_or_card(
            title=SENSITIVITY_CARDGROUP_TITLE,
            subtitle=SENSITIVITY_CARDGROUP_SUBTITLE,
            children=cards,
        )


def compute_sensitivity_adhoc(
    adapter: Adapter | None = None,
    metric_names: Sequence[str] | None = None,
    labels: Mapping[str, str] | None = None,
    order: Literal["first", "second", "total"] = "total",
    top_k: int | None = None,
) -> AnalysisCardBase:
    """
    Compute SensitivityAnalysis cards for the given experiment and either Adapter or
    GenerationStrategy.

    Note that cards are not saved to the database when computed adhoc -- they are only
    saved when computed as part of call to ``Client.compute_analyses`` or equivalent.

    Args:
        adapter: The adapter to use to compute the analysis. If not provided, will use
            the current adapter on the ``GenerationStrategy``.
        metric_names: The names of the metrics and outcomes for which to compute
                sensitivities. This should preferably be metrics with a good model fit.
                Defaults to all metrics in the experiment.
        order: A string specifying the order of the Sobol indices to be computed.
            Supports "first" and "total" and defaults to "first".
        top_k: Optional limit on the number of parameters to show in the plot.
        labels: A mapping from metric names to labels to use in the plot. If a label
            is not provided for a metric, the metric name will be used.
    """
    analysis = SensitivityAnalysisPlot(
        metric_names=metric_names,
        order=order,
        top_k=top_k,
        labels=labels,
    )
    return analysis.compute(adapter=adapter)


def _prepare_data(
    adapter: TorchAdapter,
    metric_names: Sequence[str] | None,
    order: Literal["first", "second", "total"],
) -> pd.DataFrame:
    sensitivities = ax_parameter_sens(
        adapter=adapter,
        metrics=[*metric_names] if metric_names is not None else None,
        order=order,
    )

    return pd.DataFrame.from_records(
        [
            {
                "metric_name": metric_name,
                "parameter_name": parameter_name,
                "sensitivity": sensitivity,
            }
            for metric_name, sensitivity_dict in sensitivities.items()
            for parameter_name, sensitivity in sensitivity_dict.items()
        ]
    )


def _prepare_card_components(
    data: pd.DataFrame,
    metric_name: str,
    metric_label: str,
    top_k: int | None,
) -> tuple[pd.DataFrame, go.Figure]:
    plotting_df = data.loc[data["metric_name"] == metric_name][
        ["parameter_name", "sensitivity"]
    ].copy()

    # If the parameter name is too long, truncate it.
    # If the parameter name is a second order interaction, truncate each parameter name
    # separately then re-combine.
    # If the truncated parameter name already exists, append count at end to prevent
    # collisions.
    # TODO: @paschali @mgarrard clean up after implementing parameter canonical names
    param_names = plotting_df["parameter_name"].unique()
    param_to_shortened_name = {}
    shortened_name_count = {}
    for name in param_names:
        shortened_name = (
            " & ".join(
                truncate_label(label=sub_name, n=MAX_LABEL_LEN // 2)
                for sub_name in name.split(" & ")
            )
            if "&" in name
            else truncate_label(label=name, n=MAX_LABEL_LEN)
        )
        # track number of times each shortened name is seen
        if shortened_name not in shortened_name_count:
            shortened_name_count[shortened_name] = 0
        else:
            shortened_name_count[shortened_name] += 1
            shortened_name = shortened_name + f"_{shortened_name_count[shortened_name]}"
        param_to_shortened_name[name] = shortened_name
    plotting_df["truncated_parameter_name"] = plotting_df["parameter_name"].map(
        param_to_shortened_name
    )

    plotting_df["importance"] = plotting_df["sensitivity"].abs()
    plotting_df["direction"] = plotting_df["sensitivity"].apply(
        lambda x: f"Increases {metric_label}" if x >= 0 else f"Decreases {metric_label}"
    )
    figure = px.bar(
        plotting_df.sort_values(by="importance", ascending=False)
        .reset_index()
        .head(top_k),
        x="importance",
        y="truncated_parameter_name",
        orientation="h",
        color="direction",
        color_discrete_map={
            f"Increases {metric_label}": COLOR_FOR_INCREASES,
            f"Decreases {metric_label}": COLOR_FOR_DECREASES,
        },
        # Show longer version of parameter name on hover without overflowing hover
        hover_data=["parameter_name"][:MAX_HOVER_LABEL_LEN],
    )

    figure.update_layout(
        # Display most important parameters first
        yaxis={"categoryorder": "total ascending"},
        # move legend to bottom of plot
        legend=LEGEND_POSITION,
        margin=MARGIN_REDUCUTION,
    )

    return (
        plotting_df[["parameter_name", "sensitivity"]],
        figure,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from numbers import Number
from typing import Any, TYPE_CHECKING

import numpy as np
from ax.adapter.transforms.derelativize import Derelativize
from ax.core.observation import ObservationData
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import Parameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import HierarchicalSearchSpace, RobustSearchSpace, SearchSpace
from ax.exceptions.core import UserInputError
from numpy import typing as npt
from pyre_extensions import none_throws
from scipy.stats import norm

T_MATCH_CI_WIDTH = Callable[[npt.NDArray, npt.NDArray], tuple[npt.NDArray, npt.NDArray]]


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


# pyre-fixme[24]: Generic type `dict` expects 2 type parameters, use `typing.Dict`
#  to avoid runtime subscripting errors.
class ClosestLookupDict(dict):
    """A dictionary with numeric keys that looks up the closest key."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # pyre-fixme[4]: Attribute must be annotated.
        self._keys = sorted(self.keys())

    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def __setitem__(self, key: Number, val: Any) -> None:
        if not isinstance(key, Number):
            raise ValueError("ClosestLookupDict only allows numerical keys.")
        super().__setitem__(key, val)
        # pyre-fixme[6]: For 2nd argument expected `Union[bytes, complex, float,
        #  int, generic, str]` but got `Number`.
        ipos = np.searchsorted(self._keys, key)
        self._keys.insert(ipos, key)

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def __getitem__(self, key: Number) -> Any:
        try:
            return super().__getitem__(key)
        except KeyError:
            if not self.keys():
                raise RuntimeError("ClosestLookupDict is empty.")
            # pyre-fixme[6]: For 2nd argument expected `Union[bytes, complex, float,
            #  int, generic, str]` but got `Number`.
            ipos = np.searchsorted(self._keys, key)
            if ipos == 0:
                return super().__getitem__(self._keys[0])
            elif ipos == len(self._keys):
                return super().__getitem__(self._keys[-1])
            lkey, rkey = self._keys[ipos - 1 : ipos + 1]
            if np.abs(key - lkey) <= np.abs(key - rkey):  # pyre-ignore [58]
                return super().__getitem__(lkey)
            else:
                return super().__getitem__(rkey)


def get_data(
    observation_data: list[ObservationData],
    metric_names: list[str] | None = None,
    raise_on_non_finite_data: bool = True,
) -> dict[str, list[float]]:
    """Extract all metrics if `metric_names` is None.

    Raises a value error if any data is non-finite.

    Args:
        observation_data: List of observation data.
        metric_names: List of metric names.
        raise_on_non_finite_data: If true, raises an exception on nan/inf.

    Returns:
        A dictionary mapping metric names to lists of metric values.
    """
    Ys = defaultdict(list)
    for obsd in observation_data:
        for i, m in enumerate(obsd.metric_names):
            if metric_names is None or m in metric_names:
                val = obsd.means[i]
                if raise_on_non_finite_data and (not np.isfinite(val)):
                    raise ValueError(f"Non-finite data found for metric {m}: {val}")
                Ys[m].append(val)
    return Ys


def match_ci_width(
    *,
    mean: npt.NDArray,
    sem: npt.NDArray | None,
    variance: npt.NDArray | None,
    transform: Callable[[npt.NDArray], npt.NDArray],
    level: float = 0.95,
    lower_bound: float | None = None,
    upper_bound: float | None = None,
) -> tuple[npt.NDArray, npt.NDArray]:
    """Transform the mean and update the sem / variance to match the width of
    the confidence interval. The size of the transformed confidence interval
    will be proportional to the size of the original confidence interval. If the
    mean is doubled after the transform, the size of the confidence interval will
    also be doubled.

    Args:
        mean: The mean of the posterior.
        sem: The standard error of the posterior. Only one of
            `sem` or `variance` should be provided.
        variance: The variance of the posterior. Only one of
            `sem` or `variance` should be provided.
        transform: The transform to apply to the mean.
        level: The confidence level of the confidence interval used to match the width.
        lower_bound: If given, the mean and the ci bounds (computed after first clipping
            the mean) will be clipped to this value before applying the transform.
        upper_bound: If given, the mean and the ci bounds (computed after first clipping
            the mean) will be clipped to this value before applying the transform.

    Returns:
        A tuple of the transformed mean and the new sem or variance, depending on which
        one of them was provided as the input.
    """
    if variance is not None:
        if sem is not None:
            raise UserInputError("Only one of `sem` or `variance` should be provided.")
        sem = np.sqrt(variance)
    sem = none_throws(sem)
    if lower_bound is not None or upper_bound is not None:
        mean = np.clip(a=mean, a_min=lower_bound, a_max=upper_bound)
    new_mean = transform(mean)
    if np.all(np.isnan(sem)):
        # If SEM is NaN, we don't need to transform it.
        new_sem = sem
    else:
        fac = norm.ppf(1 - (1 - level) / 2)
        d = fac * sem
        right = mean + d
        left = mean - d
        if lower_bound is not None or upper_bound is not None:
            left = np.clip(a=left, a_min=lower_bound, a_max=upper_bound)
            right = np.clip(a=right, a_min=lower_bound, a_max=upper_bound)
        width_asym = transform(right) - transform(left)
        new_sem = width_asym / (2 * fac)
    if variance is not None:
        return new_mean, new_sem**2
    return new_mean, new_sem


def construct_new_search_space(
    search_space: SearchSpace,
    parameters: list[Parameter],
    parameter_constraints: list[ParameterConstraint] | None = None,
) -> SearchSpace:
    """Construct a search space with the transformed arguments.

    If the `search_space` is a `RobustSearchSpace`, this will use its
    environmental variables and distributions, and remove the environmental
    variables from `parameters` before constructing.

    Args:
        parameters: List of transformed parameter objects.
        parameter_constraints: List of parameter constraints.

    Returns:
        The new search space instance.
    """
    new_kwargs: dict[str, Any] = {
        "parameters": parameters,
        "parameter_constraints": parameter_constraints,
    }
    if isinstance(search_space, HierarchicalSearchSpace):
        # Temporarily relax the `requires_root` flag for the new search space. This is
        # fine because this function is typically called during transforms.
        new_kwargs["requires_root"] = False

    if isinstance(search_space, RobustSearchSpace):
        env_vars = list(search_space._environmental_variables.values())
        if env_vars:
            # Add environmental variables and remove them from parameters.
            new_kwargs["environmental_variables"] = env_vars
            new_kwargs["parameters"] = [p for p in parameters if p not in env_vars]
        new_kwargs["parameter_distributions"] = search_space.parameter_distributions
        new_kwargs["num_samples"] = search_space.num_samples
    return search_space.__class__(**new_kwargs)


def derelativize_optimization_config_with_raw_status_quo(
    optimization_config: OptimizationConfig, adapter: adapter_module.base.Adapter
) -> OptimizationConfig:
    """Derelativize optimization_config using raw status-quo values"""
    tf = Derelativize(
        search_space=adapter.model_space.clone(),
        config={"use_raw_status_quo": True},
    )
    return tf.transform_optimization_config(
        optimization_config=optimization_config.clone(), adapter=adapter
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.plotly.bandit_rollout import BanditRollout
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_discrete_search_space,
    get_online_experiments,
    get_sobol,
)


class TestBanditRollout(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.experiment = Experiment(
            search_space=get_discrete_search_space(), name="test_experiment"
        )

        generator = get_sobol(search_space=self.experiment.search_space)
        for _ in range(2):
            gr = generator.gen(n=2)
            self.experiment.new_batch_trial(generator_run=gr)
            self.experiment.new_batch_trial(generator_run=gr)

    def test_compute(self) -> None:
        analysis = BanditRollout()

        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()

        card = analysis.compute(experiment=self.experiment)

        self.assertEqual(card.name, "BanditRollout")
        self.assertEqual(
            card.title, f"Bandit Rollout Weights by Trial for {self.experiment.name}"
        )
        self.assertEqual(
            card.subtitle,
            (
                "The Bandit Rollout visualization provides a comprehensive "
                "overview of the allocation of weights across different trials "
                "and arms. By representing each trial as a distinct axis, this "
                "plot allows for the examination of exploration and exploitation "
                "dynamics over time. It aids in identifying trends and patterns in "
                "arm performance, offering insights into the effectiveness of the "
                "bandit algorithm. Observing the distribution of weights can "
                "reveal correlations and interactions that contribute to the "
                "success or failure of various strategies, enhancing the "
                "understanding of experimental results."
            ),
        )
        self.assertEqual(
            {*card.df.columns},
            {"trial_index", "arm_name", "arm_weight", "normalized_weight"},
        )
        self.assertIsNotNone(card.blob)

    def test_online(self) -> None:
        # Test ParallelCoordinatesPlot can be computed for a variety of experiments
        # which resemble those we see in an online setting.
        for experiment in get_online_experiments():
            analysis = BanditRollout()
            _ = analysis.compute(experiment=experiment)
